# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""Tests for JSON message streams and channels.
"""

import collections
import functools
import io
import pytest
import random
import re
import socket
import threading
import time

from debugpy.common import json, log, messaging
from tests.patterns import some


# Default timeout for tests in this file.
pytestmark = pytest.mark.timeout(5)


class JsonMemoryStream(object):
    """Like JsonIOStream, but working directly with values stored in memory.
    Values are round-tripped through JSON serialization.

    For input, values are read from the supplied sequence or iterator.
    For output, values are appended to the supplied collection.
    """

    json_decoder_factory = messaging.JsonIOStream.json_decoder_factory
    json_encoder_factory = messaging.JsonIOStream.json_encoder_factory

    def __init__(self, input, output, name="memory"):
        self.name = name
        self.input = iter(input)
        self.output = output

    def close(self):
        pass

    def _log_message(self, dir, data):
        format_string = "{0} {1} " + (
            "{2:indent=None}" if isinstance(data, list) else "{2}"
        )
        return log.debug(format_string, self.name, dir, json.repr(data))

    def read_json(self, decoder=None):
        decoder = decoder if decoder is not None else self.json_decoder_factory()
        try:
            value = next(self.input)
        except StopIteration:
            raise messaging.NoMoreMessages(stream=self)
        value = decoder.decode(json.dumps(value))
        self._log_message("-->", value)
        return value

    def write_json(self, value, encoder=None):
        encoder = encoder if encoder is not None else self.json_encoder_factory()
        value = json.loads(encoder.encode(value))
        self._log_message("<--", value)
        self.output.append(value)


class TestJsonIOStream(object):
    MESSAGE_BODY_TEMPLATE = '{"arguments": {"threadId": 3}, "command": "next", "seq": %d, "type": "request"}'
    MESSAGES = []
    SERIALIZED_MESSAGES = b""

    @classmethod
    def setup_class(cls):
        for seq in range(0, 3):
            message_body = cls.MESSAGE_BODY_TEMPLATE % seq
            message = json.loads(
                message_body, object_pairs_hook=collections.OrderedDict
            )
            message_body = message_body.encode("utf-8")
            cls.MESSAGES.append(message)
            message_header = "Content-Length: %d\r\n\r\n" % len(message_body)
            cls.SERIALIZED_MESSAGES += message_header.encode("ascii") + message_body

    def test_read(self):
        data = io.BytesIO(self.SERIALIZED_MESSAGES)
        stream = messaging.JsonIOStream(data, data, "data")
        for expected_message in self.MESSAGES:
            message = stream.read_json()
            assert message == expected_message
        with pytest.raises(messaging.NoMoreMessages) as exc_info:
            stream.read_json()
        assert exc_info.value.stream is stream

    def test_write(self):
        data = io.BytesIO()
        stream = messaging.JsonIOStream(data, data, "data")
        for message in self.MESSAGES:
            stream.write_json(message)
        data = data.getvalue()
        assert data == self.SERIALIZED_MESSAGES


class TestJsonMemoryStream(object):
    MESSAGES = [
        {"seq": 1, "type": "request", "command": "next", "arguments": {"threadId": 3}},
        {"seq": 2, "type": "request", "command": "next", "arguments": {"threadId": 5}},
    ]

    def test_read(self):
        stream = JsonMemoryStream(self.MESSAGES, [])
        for expected_message in self.MESSAGES:
            message = stream.read_json()
            assert message == expected_message
        with pytest.raises(messaging.NoMoreMessages) as exc_info:
            stream.read_json()
        assert exc_info.value.stream is stream

    def test_write(self):
        messages = []
        stream = JsonMemoryStream([], messages)
        for message in self.MESSAGES:
            stream.write_json(message)
        assert messages == self.MESSAGES


class MessageHandlerRecorder(list):
    def __call__(self, handler):
        @functools.wraps(handler)
        def record_and_handle(instance, message):
            name = handler.__name__
            if isinstance(name, bytes):
                name = name.decode("utf-8")
            record = {"channel": message.channel, "handler": name}

            if isinstance(message, messaging.Event):
                record.update(
                    {"type": "event", "event": message.event, "body": message.body}
                )
            elif isinstance(message, messaging.Request):
                record.update(
                    {
                        "type": "request",
                        "command": message.command,
                        "arguments": message.arguments,
                    }
                )

            self.append(record)
            return handler(instance, message)

        return record_and_handle

    def expect(self, channel, inputs, handlers):
        expected_records = []
        for input, handler in zip(inputs, handlers):
            expected_record = {"channel": channel, "handler": handler}
            expected_record.update(
                {
                    key: value
                    for key, value in input.items()
                    if key in ("type", "event", "command", "body", "arguments")
                }
            )
            expected_records.append(expected_record)
        assert expected_records == self


class TestJsonMessageChannel(object):
    @staticmethod
    def iter_with_event(collection):
        """Like iter(), but also exposes a threading.Event that is set
        when the returned iterator is exhausted.
        """
        exhausted = threading.Event()

        def iterate():
            for x in collection:
                yield x
            exhausted.set()

        return iterate(), exhausted

    def test_events(self):
        EVENTS = [
            {
                "seq": 1,
                "type": "event",
                "event": "stopped",
                "body": {"reason": "pause"},
            },
            {
                "seq": 2,
                "type": "event",
                "event": "unknown",
                "body": {"something": "else"},
            },
        ]

        recorder = MessageHandlerRecorder()

        class Handlers(object):
            @recorder
            def stopped_event(self, event):
                assert event.event == "stopped"

            @recorder
            def event(self, event):
                assert event.event == "unknown"

        stream = JsonMemoryStream(EVENTS, [])
        channel = messaging.JsonMessageChannel(stream, Handlers())
        channel.start()
        channel.wait()

        recorder.expect(channel, EVENTS, ["stopped_event", "event"])

    def test_requests(self):
        REQUESTS = [
            {
                "seq": 1,
                "type": "request",
                "command": "next",
                "arguments": {"threadId": 3},
            },
            {
                "seq": 2,
                "type": "request",
                "command": "launch",
                "arguments": {"program": "main.py"},
            },
            {
                "seq": 3,
                "type": "request",
                "command": "unknown",
                "arguments": {"answer": 42},
            },
            {
                "seq": 4,
                "type": "request",
                "command": "pause",
                "arguments": {"threadId": 5},
            },
        ]

        recorder = MessageHandlerRecorder()

        class Handlers(object):
            @recorder
            def next_request(self, request):
                assert request.command == "next"
                return {"threadId": 7}

            @recorder
            def launch_request(self, request):
                assert request.command == "launch"
                self._launch = request
                return messaging.NO_RESPONSE

            @recorder
            def request(self, request):
                request.respond({})

            @recorder
            def pause_request(self, request):
                assert request.command == "pause"
                self._launch.respond({"processId": 9})
                raise request.cant_handle("pause error")

        stream = JsonMemoryStream(REQUESTS, [])
        channel = messaging.JsonMessageChannel(stream, Handlers())
        channel.start()
        channel.wait()

        recorder.expect(
            channel,
            REQUESTS,
            ["next_request", "launch_request", "request", "pause_request"],
        )

        assert stream.output == [
            {
                "seq": 1,
                "type": "response",
                "request_seq": 1,
                "command": "next",
                "success": True,
                "body": {"threadId": 7},
            },
            {
                "seq": 2,
                "type": "response",
                "request_seq": 3,
                "command": "unknown",
                "success": True,
            },
            {
                "seq": 3,
                "type": "response",
                "request_seq": 2,
                "command": "launch",
                "success": True,
                "body": {"processId": 9},
            },
            {
                "seq": 4,
                "type": "response",
                "request_seq": 4,
                "command": "pause",
                "success": False,
                "message": "pause error",
            },
        ]

    def test_responses(self):
        request1_sent = threading.Event()
        request2_sent = threading.Event()
        request3_sent = threading.Event()
        request4_sent = threading.Event()

        def iter_responses():
            request1_sent.wait()
            yield {
                "seq": 1,
                "type": "response",
                "request_seq": 1,
                "command": "next",
                "success": True,
                "body": {"threadId": 3},
            }

            request2_sent.wait()
            yield {
                "seq": 2,
                "type": "response",
                "request_seq": 2,
                "command": "pause",
                "success": False,
                "message": "Invalid message: pause not supported",
            }

            request3_sent.wait()
            yield {
                "seq": 3,
                "type": "response",
                "request_seq": 3,
                "command": "next",
                "success": True,
                "body": {"threadId": 5},
            }

            request4_sent.wait()

        stream = JsonMemoryStream(iter_responses(), [])
        channel = messaging.JsonMessageChannel(stream, None)
        channel.start()

        # Blocking wait.
        request1 = channel.send_request("next")
        request1_sent.set()
        log.info("Waiting for response...")
        response1_body = request1.wait_for_response()
        response1 = request1.response

        assert response1.success
        assert response1.request is request1
        assert response1.body == response1_body
        assert response1.body == {"threadId": 3}

        # Async callback, registered before response is received.
        request2 = channel.send_request("pause")
        response2 = []
        response2_received = threading.Event()

        def response2_handler(resp):
            response2.append(resp)
            response2_received.set()

        log.info("Registering callback")
        request2.on_response(response2_handler)
        request2_sent.set()

        log.info("Waiting for callback...")
        response2_received.wait()
        (response2,) = response2

        assert not response2.success
        assert response2.request is request2
        assert response2 is request2.response
        assert response2.body == messaging.InvalidMessageError(
            "pause not supported", request2
        )

        # Async callback, registered after response is received.
        request3 = channel.send_request("next")
        request3_sent.set()
        request3.wait_for_response()
        response3 = []
        response3_received = threading.Event()

        def response3_handler(resp):
            response3.append(resp)
            response3_received.set()

        log.info("Registering callback")
        request3.on_response(response3_handler)

        log.info("Waiting for callback...")
        response3_received.wait()
        (response3,) = response3

        assert response3.success
        assert response3.request is request3
        assert response3 is request3.response
        assert response3.body == {"threadId": 5}

        # Async callback, registered after channel is closed.
        request4 = channel.send_request("next")
        request4_sent.set()
        channel.wait()
        response4 = []
        response4_received = threading.Event()

        def response4_handler(resp):
            response4.append(resp)
            response4_received.set()

        log.info("Registering callback")
        request4.on_response(response4_handler)

        log.info("Waiting for callback...")
        response4_received.wait()
        (response4,) = response4

        assert not response4.success
        assert response4.request is request4
        assert response4 is request4.response
        assert isinstance(response4.body, messaging.NoMoreMessages)

    def test_invalid_request_handling(self):
        REQUESTS = [
            {
                "seq": 1,
                "type": "request",
                "command": "stackTrace",
                "arguments": {"AAA": {}},
            },
            {"seq": 2, "type": "request", "command": "stackTrace", "arguments": {}},
            {"seq": 3, "type": "request", "command": "unknown", "arguments": None},
            {"seq": 4, "type": "request", "command": "pause"},
        ]

        class Handlers(object):
            def stackTrace_request(self, request):
                request.arguments["AAA"]
                request.arguments["AAA"]["BBB"]

            def request(self, request):
                request.arguments["CCC"]

            def pause_request(self, request):
                request.arguments["DDD"]

        output = []
        stream = JsonMemoryStream(REQUESTS, output)
        channel = messaging.JsonMessageChannel(stream, Handlers())
        channel.start()
        channel.wait()

        def missing_property(name):
            return some.str.matching("Invalid message:.*" + re.escape(name) + ".*")

        assert output == [
            {
                "seq": 1,
                "type": "response",
                "request_seq": 1,
                "command": "stackTrace",
                "success": False,
                "message": missing_property("BBB"),
            },
            {
                "seq": 2,
                "type": "response",
                "request_seq": 2,
                "command": "stackTrace",
                "success": False,
                "message": missing_property("AAA"),
            },
            {
                "seq": 3,
                "type": "response",
                "request_seq": 3,
                "command": "unknown",
                "success": False,
                "message": missing_property("CCC"),
            },
            {
                "seq": 4,
                "type": "response",
                "request_seq": 4,
                "command": "pause",
                "success": False,
                "message": missing_property("DDD"),
            },
        ]

    def test_fuzz(self):
        # Set up two channels over the same stream that send messages to each other
        # asynchronously, and record everything that they send and receive.
        # All records should match at the end.

        class Fuzzer(object):
            def __init__(self, name):
                self.name = name
                self.lock = threading.Lock()
                self.sent = []
                self.received = []
                self.responses_sent = []
                self.responses_received = []
                self.done = False

            def start(self, channel):
                self._worker = threading.Thread(
                    name=self.name,
                    target=lambda: self._send_requests_and_events(channel),
                )
                self._worker.daemon = True
                self._worker.start()

            def wait(self):
                self._worker.join()

            def done_event(self, event):
                with self.lock:
                    self.done = True

            def fizz_event(self, event):
                assert event.event == "fizz"
                with self.lock:
                    self.received.append(("event", "fizz", event.body))

            def buzz_event(self, event):
                assert event.event == "buzz"
                with self.lock:
                    self.received.append(("event", "buzz", event.body))

            def event(self, event):
                with self.lock:
                    self.received.append(("event", event.event, event.body))

            def make_and_log_response(self, request):
                x = random.randint(-100, 100)
                if x < 0:
                    exc_type = (
                        messaging.InvalidMessageError
                        if x % 2
                        else messaging.MessageHandlingError
                    )
                    x = exc_type(str(x), request)
                with self.lock:
                    self.responses_sent.append((request.seq, x))
                return x

            def fizz_request(self, request):
                assert request.command == "fizz"
                with self.lock:
                    self.received.append(("request", "fizz", request.arguments))
                return self.make_and_log_response(request)

            def buzz_request(self, request):
                assert request.command == "buzz"
                with self.lock:
                    self.received.append(("request", "buzz", request.arguments))
                return self.make_and_log_response(request)

            def request(self, request):
                with self.lock:
                    self.received.append(
                        ("request", request.command, request.arguments)
                    )
                return self.make_and_log_response(request)

            def _got_response(self, response):
                with self.lock:
                    self.responses_received.append(
                        (response.request.seq, response.body)
                    )

            def _send_requests_and_events(self, channel):
                types = [random.choice(("event", "request")) for _ in range(0, 100)]

                for typ in types:
                    name = random.choice(("fizz", "buzz", "fizzbuzz"))
                    body = random.randint(0, 100)

                    with self.lock:
                        self.sent.append((typ, name, body))

                    if typ == "event":
                        channel.send_event(name, body)
                    elif typ == "request":
                        req = channel.send_request(name, body)
                        req.on_response(self._got_response)

                channel.send_event("done")

                # Spin until we receive "done", and also get responses to all requests.
                requests_sent = types.count("request")
                log.info("{0} waiting for {1} responses...", self.name, requests_sent)
                while True:
                    with self.lock:
                        if self.done:
                            if requests_sent == len(self.responses_received):
                                break
                    time.sleep(0.1)

        fuzzer1 = Fuzzer("fuzzer1")
        fuzzer2 = Fuzzer("fuzzer2")

        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.bind(("localhost", 0))
        _, port = server_socket.getsockname()
        server_socket.listen(0)

        socket1 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket1_thread = threading.Thread(
            target=lambda: socket1.connect(("localhost", port))
        )
        socket1_thread.start()
        socket2, _ = server_socket.accept()
        socket1_thread.join()

        try:
            io1 = socket1.makefile("rwb", 0)
            io2 = socket2.makefile("rwb", 0)

            stream1 = messaging.JsonIOStream(io1, io1, "socket1")
            channel1 = messaging.JsonMessageChannel(stream1, fuzzer1)
            channel1.start()
            fuzzer1.start(channel1)

            stream2 = messaging.JsonIOStream(io2, io2, "socket2")
            channel2 = messaging.JsonMessageChannel(stream2, fuzzer2)
            channel2.start()
            fuzzer2.start(channel2)

            fuzzer1.wait()
            fuzzer2.wait()

        finally:
            socket1.close()
            socket2.close()

        assert fuzzer1.sent == fuzzer2.received
        assert fuzzer2.sent == fuzzer1.received
        assert fuzzer1.responses_sent == fuzzer2.responses_received
        assert fuzzer2.responses_sent == fuzzer1.responses_received


class TestTypeConversion(object):
    def test_str_to_num(self):

        # test conversion that are expected to work
        correct_trials = [("1.0", float), ("1", int), ("1", bool)]
        for val_trial, type_trial in correct_trials:
            assert isinstance(
                json.of_type(type_trial)(val_trial), type_trial
            ), "Wrong type coversion"

        # test conversion that are not expected to work
        try:
            json.of_type(int)("1.0")
            raise ValueError("This test should have failed")
        except TypeError:
            pass



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import os
import psutil
import pytest
import sys

import debugpy
import tests
from tests import debug, log, timeline
from tests.debug import runners
from tests.patterns import some


if not tests.full:

    @pytest.fixture(params=[runners.launch] + runners.all_attach_socket)
    def run(request):
        return request.param


def expected_subprocess_config(parent_session):
    config = dict(parent_session.config)
    for key in "args", "listen", "postDebugTask", "preLaunchTask", "processId":
        config.pop(key, None)
    for key in "python", "pythonArgs", "pythonPath":
        if key in config:
            config[key] = some.thing
    config.update(
        {
            "name": some.str,
            "request": "attach",
            "subProcessId": some.int,
            "connect": {"host": some.str, "port": some.int},
        }
    )
    return config


@pytest.mark.parametrize(
    "start_method",
    ["spawn"] if sys.platform == "win32" else ["spawn", "fork"],
)
def test_multiprocessing(pyfile, target, run, start_method):
    if start_method == "spawn" and sys.platform != "win32":
        pytest.skip("https://github.com/microsoft/ptvsd/issues/1887")

    @pyfile
    def code_to_debug():
        import debuggee
        import multiprocessing
        import os
        import sys

        # https://github.com/microsoft/ptvsd/issues/2108
        class Foo(object):
            pass

        def parent(q, a):
            from debuggee import backchannel

            debuggee.setup()

            print("spawning child")
            p = multiprocessing.Process(target=child, args=(q, a))
            p.start()
            print("child spawned")

            q.put("foo?")
            foo = a.get()
            assert isinstance(foo, Foo), repr(foo)

            q.put("child_pid?")
            what, child_pid = a.get()
            assert what == "child_pid"
            backchannel.send(child_pid)

            q.put("grandchild_pid?")
            what, grandchild_pid = a.get()
            assert what == "grandchild_pid"
            backchannel.send(grandchild_pid)

            assert backchannel.receive() == "continue"
            q.put("exit!")
            p.join()

        def child(q, a):
            print("entering child")
            assert q.get() == "foo?"
            a.put(Foo())

            assert q.get() == "child_pid?"
            a.put(("child_pid", os.getpid()))

            print("spawning child of child")
            p = multiprocessing.Process(target=grandchild, args=(q, a))
            p.start()
            p.join()

            print("leaving child")

        def grandchild(q, a):
            print("entering grandchild")
            assert q.get() == "grandchild_pid?"
            a.put(("grandchild_pid", os.getpid()))

            assert q.get() == "exit!"
            print("leaving grandchild")

        if __name__ == "__main__":
            start_method = sys.argv[1]
            if start_method != "":
                multiprocessing.set_start_method(start_method)

            q = multiprocessing.Queue()
            a = multiprocessing.Queue()
            try:
                parent(q, a)
            finally:
                q.close()
                a.close()

    with debug.Session() as parent_session:
        parent_backchannel = parent_session.open_backchannel()

        with run(parent_session, target(code_to_debug, args=[start_method])):
            pass

        expected_child_config = expected_subprocess_config(parent_session)
        child_config = parent_session.wait_for_next_event("debugpyAttach")
        child_config.pop("isOutputRedirected", None)
        assert child_config == expected_child_config
        parent_session.proceed()

        with debug.Session(child_config) as child_session:
            with child_session.start():
                pass

            expected_grandchild_config = expected_subprocess_config(child_session)
            grandchild_config = child_session.wait_for_next_event("debugpyAttach")
            grandchild_config.pop("isOutputRedirected", None)
            assert grandchild_config == expected_grandchild_config

            with debug.Session(grandchild_config) as grandchild_session:
                with grandchild_session.start():
                    pass

                parent_backchannel.send("continue")


@pytest.mark.parametrize("subProcess", [True, False, None])
@pytest.mark.parametrize("method", ["startDebugging", "debugpyAttach", ""])
def test_subprocess(pyfile, target, run, subProcess, method):
    @pyfile
    def child():
        import os
        import sys

        assert "debugpy" in sys.modules

        import debugpy
        from debuggee import backchannel

        backchannel.send(os.getpid())
        backchannel.send(debugpy.__file__)
        backchannel.send(sys.argv)

    @pyfile
    def parent():
        import debuggee
        import os
        import subprocess
        import sys

        debuggee.setup()
        argv = [sys.executable, sys.argv[1], "--arg1", "--arg2", "--arg3"]
        env = os.environ.copy()
        process = subprocess.Popen(
            argv,
            env=env,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        process.wait()

    with debug.Session() as parent_session:
        backchannel = parent_session.open_backchannel()

        if method:
            parent_session.capabilities["supportsStartDebuggingRequest"] = (method == "startDebugging")
        parent_session.config["preLaunchTask"] = "doSomething"
        parent_session.config["postDebugTask"] = "doSomethingElse"
        if subProcess is not None:
            parent_session.config["subProcess"] = subProcess

        with run(parent_session, target(parent, args=[child])):
            pass

        if subProcess is False:
            return

        expected_child_config = expected_subprocess_config(parent_session)
        
        if method == "startDebugging":
            subprocess_request = parent_session.timeline.wait_for_next(timeline.Request("startDebugging"))
            child_config = subprocess_request.arguments("configuration", dict)
            del expected_child_config["request"]
        else:
            child_config = parent_session.wait_for_next_event("debugpyAttach")

        child_config = dict(child_config)
        child_config.pop("isOutputRedirected", None)
        assert child_config == expected_child_config
        child_config["request"] = "attach"

        parent_session.proceed()

        with debug.Session(child_config) as child_session:
            with child_session.start():
                pass

            child_pid = backchannel.receive()
            assert child_pid == child_config["subProcessId"]
            assert str(child_pid) in child_config["name"]

            debugpy_file = backchannel.receive()
            assert os.path.abspath(debugpy_file) == os.path.abspath(debugpy.__file__)

            child_argv = backchannel.receive()
            assert child_argv == [child, "--arg1", "--arg2", "--arg3"]


@pytest.mark.parametrize("run", runners.all_launch)
def test_autokill(pyfile, target, run):
    @pyfile
    def child():
        import os
        from debuggee import backchannel

        backchannel.send(os.getpid())
        while True:
            pass

    @pyfile
    def parent():
        import debuggee
        import os
        import subprocess
        import sys

        debuggee.setup()
        argv = [sys.executable, sys.argv[1]]
        env = os.environ.copy()
        subprocess.Popen(
            argv,
            env=env,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ).wait()

    with debug.Session() as parent_session:
        parent_session.expected_exit_code = some.int

        backchannel = parent_session.open_backchannel()
        with run(parent_session, target(parent, args=[child])):
            pass

        child_config = parent_session.wait_for_next_event("debugpyAttach")
        parent_session.proceed()

        with debug.Session(child_config) as child_session:
            with child_session.start():
                pass

            child_pid = backchannel.receive()
            assert child_config["subProcessId"] == child_pid
            child_process = psutil.Process(child_pid)

            parent_session.request("terminate")
            child_session.wait_for_exit()

    log.info("Waiting for child process...")
    child_process.wait()


@pytest.mark.parametrize("run", runners.all_launch)
def test_autokill_nodebug(pyfile, target, run):
    @pyfile
    def child():
        import os
        from debuggee import backchannel

        backchannel.send(os.getpid())
        while True:
            pass

    @pyfile
    def parent():
        import os
        import subprocess
        import sys

        argv = [sys.executable, sys.argv[1]]
        env = os.environ.copy()
        subprocess.Popen(
            argv,
            env=env,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ).wait()

    with debug.Session() as session:
        session.expected_exit_code = some.int
        session.config["noDebug"] = True

        backchannel = session.open_backchannel()
        run(session, target(parent, args=[child]))

        child_pid = backchannel.receive()
        child_process = psutil.Process(child_pid)

        session.request("terminate")

    log.info("Waiting for child process...")
    child_process.wait()


def test_argv_quoting(pyfile, target, run):
    @pyfile
    def args():
        args = [  # noqa
            r"regular",
            r"",
            r"with spaces" r'"quoted"',
            r'" quote at start',
            r'quote at end "',
            r'quote in " the middle',
            r'quotes "in the" middle',
            r"\path with\spaces",
            r"\path\with\terminal\backslash" + "\\",
            r"backslash \" before quote",
        ]

    @pyfile
    def parent():
        import debuggee
        import sys
        import subprocess
        from args import args

        debuggee.setup()
        child = sys.argv[1]
        subprocess.check_call([sys.executable] + [child] + args)

    @pyfile
    def child():
        import sys
        from debuggee import backchannel
        from args import args as expected_args

        backchannel.send(expected_args)

        actual_args = sys.argv[1:]
        backchannel.send(actual_args)

    with debug.Session() as parent_session:
        backchannel = parent_session.open_backchannel()

        with run(parent_session, target(parent, args=[child])):
            pass

        child_config = parent_session.wait_for_next_event("debugpyAttach")
        parent_session.proceed()

        with debug.Session(child_config) as child_session:
            with child_session.start():
                pass

            expected_args = backchannel.receive()
            actual_args = backchannel.receive()

            assert expected_args == actual_args


def test_echo_and_shell(pyfile, target, run):
    """
    Checks https://github.com/microsoft/ptvsd/issues/1548
    """

    @pyfile
    def code_to_run():
        import debuggee
        import sys
        import subprocess
        import os

        debuggee.setup()

        if sys.platform == "win32":
            args = ["dir", "-c", "."]
        else:
            args = ["ls", "-c", "-la"]

        p = subprocess.Popen(
            args,
            shell=True,
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
            cwd=os.path.dirname(os.path.abspath(__file__)),
        )
        stdout, _stderr = p.communicate()
        stdout = stdout.decode("utf-8")

        if "code_to_run.py" not in stdout:
            raise AssertionError(
                'Did not find "code_to_run.py" when listing this dir with subprocess. Contents: %s'
                % (stdout,)
            )

    with debug.Session() as parent_session:
        with run(parent_session, target(code_to_run)):
            pass


@pytest.mark.parametrize("run", runners.all_attach_connect)
@pytest.mark.parametrize("wait", ["wait", ""])
def test_subprocess_unobserved(pyfile, run, target, wait):
    @pyfile
    def child():
        from debuggee import backchannel  # @bp

        backchannel.send("child running")
        backchannel.receive()

    @pyfile
    def parent():
        import debuggee
        import os
        import subprocess
        import sys

        debuggee.setup()
        args = [sys.executable, sys.argv[1]]
        env = os.environ.copy()
        subprocess.Popen(args, env=env).wait()

    with debug.Session() as parent_session:
        backchannel = parent_session.open_backchannel()

        if not wait:
            # The child process should have started running user code as soon as it's
            # spawned, before there is a client connection.
            def before_connect(address):
                assert backchannel.receive() == "child running"

            parent_session.before_connect = before_connect

        with run(parent_session, target(parent, args=[child]), wait=bool(wait)):
            pass

        child_config = parent_session.wait_for_next_event("debugpyAttach")
        parent_session.proceed()

        with debug.Session(child_config) as child_session:
            with child_session.start():
                child_session.set_breakpoints(child, all)

            if wait:
                # The child process should not have started running user code until
                # there was a client connection, so the breakpoint should be hit.
                child_session.wait_for_stop(
                    expected_frames=[some.dap.frame(child, line="bp")]
                )
                child_session.request_continue()
            else:
                # The breakpoint shouldn't be hit, since that line should have been
                # executed before we attached.
                pass

            backchannel.send("proceed")
            child_session.wait_for_terminated()


@pytest.mark.parametrize("run", runners.all_launch)
@pytest.mark.skipif(
    sys.platform != "win32", reason="job objects are specific to Windows"
)
def test_breakaway_job(pyfile, target, run):
    @pyfile
    def child():
        import os
        from debuggee import backchannel

        backchannel.send(os.getpid())
        backchannel.receive()
        backchannel.send("ok")

    @pyfile
    def parent():
        import debuggee
        import os
        import subprocess
        import sys

        debuggee.setup()
        argv = [sys.executable, sys.argv[1]]
        env = os.environ.copy()
        pipe_in = open(os.devnull, "r")
        pipe_out = open(os.devnull, "w")
        CREATE_BREAKAWAY_FROM_JOB = 0x01000000
        proc = subprocess.Popen(
            argv,
            env=env,
            stdin=pipe_in,
            stdout=pipe_out,
            stderr=pipe_out,
            creationflags=CREATE_BREAKAWAY_FROM_JOB,
        )
        pipe_in.close()
        pipe_out.close()
        proc.wait()

    with debug.Session() as parent_session:
        parent_session.config.update(
            {
                "redirectOutput": False,
                "subProcess": False,
            }
        )
        parent_session.expected_exit_code = some.int
        backchannel = parent_session.open_backchannel()

        with run(parent_session, target(parent, args=[child])):
            pass

        child_pid = backchannel.receive()
        child_process = psutil.Process(child_pid)

        parent_session.request("terminate")
        parent_session.wait_for_exit()

        # child should still be running
        backchannel.send("proceed")
        assert backchannel.receive() == "ok"

    log.info("Waiting for child process...")
    child_process.wait()


@pytest.mark.parametrize("run", runners.all_launch)
@pytest.mark.skipif(
    sys.platform == "win32", reason="os.exec() is specific to POSIX"
)
def test_subprocess_replace(pyfile, target, run):
    @pyfile
    def child():
        import os
        import sys

        assert "debugpy" in sys.modules

        from debuggee import backchannel

        backchannel.send(os.getpid())

    @pyfile
    def parent():
        import debuggee
        import os
        import sys

        debuggee.setup()
        print(f"execl({sys.executable!r}, {sys.argv[1]!r})")
        os.execl(sys.executable, sys.executable, sys.argv[1])

    with debug.Session() as parent_session:
        backchannel = parent_session.open_backchannel()
        with run(parent_session, target(parent, args=[child])):
            pass

        expected_child_config = expected_subprocess_config(parent_session)
        child_config = parent_session.wait_for_next_event("debugpyAttach")
        child_config.pop("isOutputRedirected", None)
        assert child_config == expected_child_config
        parent_session.proceed()

        with debug.Session(child_config) as child_session:
            with child_session.start():
                pass

            child_pid = backchannel.receive()
            assert child_pid == child_config["subProcessId"]
            assert str(child_pid) in child_config["name"]



#!/usr/bin/env python

# pyright: strict

import argparse
import os
import shlex
import subprocess
import sys
from contextlib import contextmanager
from typing import Iterator

_SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))

os.chdir(_SCRIPT_DIR)

_GIT_SUBREPO_ROOT = os.path.join(_SCRIPT_DIR, "build", "git-subrepo")
os.environ["GIT_SUBREPO_ROOT"] = _GIT_SUBREPO_ROOT
os.environ["PATH"] = (
    os.path.join(_GIT_SUBREPO_ROOT, "lib") + os.pathsep + os.environ["PATH"]
)
os.environ["FILTER_BRANCH_SQUELCH_WARNING"] = "1"

_GIT_URL = "https://github.com/fabioz/PyDev.Debugger.git"
_SUBREPO_NAME = "src/debugpy/_vendored/pydevd"
_SUBREPO_TMP = ".git/tmp/subrepo/" + _SUBREPO_NAME


@contextmanager
def cwd(p: str) -> Iterator[None]:
    old = os.getcwd()
    os.chdir(p)
    try:
        yield
    finally:
        os.chdir(old)


def invoke_call(*args: str) -> None:
    print(f"== {shlex.join(args)} ==")
    subprocess.check_call(args)


def invoke_call_ok(*args: str) -> bool:
    try:
        subprocess.check_call(args)
        return True
    except:
        return False


def invoke_output(*args: str, no_log: bool = False) -> str:
    if not no_log:
        print(f"== {shlex.join(args)} ==")

    return subprocess.check_output(args, text=True)


def get_current_commit() -> str:
    return invoke_output(
        "git",
        "config",
        "--file",
        f"{_SUBREPO_NAME}/.gitrepo",
        "subrepo.commit",
        no_log=True,
    ).strip()


def err_exit(message: str):
    print(message, file=sys.stderr)
    sys.exit(1)


def clone() -> None:
    # Clone the repo.
    invoke_call("git", "subrepo", "clone", _GIT_URL, _SUBREPO_NAME)


def reclone() -> None:
    # Remove the temporary branch and worktree.
    invoke_call("git", "subrepo", "clean", _SUBREPO_NAME)
    # Force clone the repo.
    invoke_call("git", "subrepo", "clone", "--force", _GIT_URL, _SUBREPO_NAME)


def pull() -> None:
    # Remove the temporary branch and worktree.
    invoke_call("git", "subrepo", "clean", "--force", _SUBREPO_NAME)

    invoke_call("git", "subrepo", "fetch", _SUBREPO_NAME)
    with cwd(_SUBREPO_NAME):
        new_commit = invoke_output("git", "rev-parse","--verify", "FETCH_HEAD").strip()

    print(f"Updating to pydevd commit {new_commit}.")

    # Pull changes and squash commit them.
    invoke_call("git", "subrepo", "pull", _SUBREPO_NAME)

    # Now, branch to see if there are any diffs. If not, then we can just reclone.
    print("Branching to check if the pydevd tree is clean.")
    invoke_call("git", "subrepo", "clean", _SUBREPO_NAME)
    invoke_call("git", "subrepo", "branch", _SUBREPO_NAME)

    with cwd(_SUBREPO_TMP):
        no_diff = invoke_call_ok("git", "diff", "--quiet", new_commit)

    if no_diff:
        # No diff, so it's safe to manually move the subrepo parent to HEAD.
        print("pydevd tree is clean, moving subrepo parent.")
        new_parent = invoke_output("git", "rev-parse", "HEAD", no_log=True).strip()
        invoke_call(
            "git",
            "config",
            "--file",
            "src/debugpy/_vendored/pydevd/.gitrepo",
            "subrepo.parent",
            new_parent,
        )
        invoke_call("git", "commit", "-am", "Update git-subrepo parent")
    else:
        print("pydevd tree has changes not pushed upstream.")


def branch(message: str) -> None:
    current_commit = get_current_commit()

    # Remove the temporary branch and worktree.
    invoke_call("git", "subrepo", "clean", _SUBREPO_NAME)
    # Ensure we have all of the subrepo refs.
    invoke_call("git", "subrepo", "fetch", _SUBREPO_NAME)
    # Populate the subrepo/src/debugpy/_vendored/pydevd branch with new changes.
    invoke_call("git", "subrepo", "branch", _SUBREPO_NAME)

    # Enter worktree; changes here are applied to the subrepo/src/debugpy/_vendored/pydevd branch.
    with cwd(_SUBREPO_TMP):
        # Remove all commits after the last pull and restage the difference.
        invoke_call("git", "reset", "--soft", current_commit)
        
        hasModifiedFile = invoke_output("git", "status","-s").strip()

        if hasModifiedFile:
            # Commit changes with a new message.
            invoke_call("git", "commit", "-m", message)


def push_to_fork(fork_remote: str, fork_branch: str) -> None:
    invoke_call("git", "push", fork_remote, f"subrepo/src/debugpy/_vendored/pydevd:{fork_branch}")


def commit() -> None:
    invoke_call("git", "subrepo", "commit", _SUBREPO_NAME)


def main() -> None:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    subparsers.add_parser(
        "clone",
        help="clones pydevd for the first time",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers.add_parser(
        "reclone",
        help="force reclones pydevd",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers.add_parser(
        "pull",
        help="pulls pydevd",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    branch_parser = subparsers.add_parser(
        "branch",
        help="squashes changes to pydevd",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    branch_parser.add_argument(
        "-m",
        "--message",
        dest="message",
        required=True,
        help="message for the squashed commit",
    )

    ptf_parser = subparsers.add_parser(
        "push-to-fork",
        help="pushes the squashed changes to pydevd",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    ptf_parser.add_argument(
        "--fork-remote",
        dest="fork_remote",
        default="pydevd-fork",
        help="pydevd remote",
    )
    ptf_parser.add_argument(
        "--fork-branch",
        dest="fork_branch",
        required=True,
        help="branch to push to on the remote",
    )

    subparsers.add_parser(
        "commit",
        help=f"runs 'git subrepo commit {_SUBREPO_NAME}'",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    args = parser.parse_args()

    if args.subcommand == "clone":
        clone()
    elif args.subcommand == "reclone":
        reclone()
    elif args.subcommand == "pull":
        pull()
    elif args.subcommand == "branch":
        branch(args.message)
    elif args.subcommand == "push-to-fork":
        push_to_fork(args.fork_remote, args.fork_branch)
    elif args.subcommand == "commit":
        commit()
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



def test_dump_threads():
    import pydevd
    from io import StringIO

    stream = StringIO()
    pydevd.dump_threads(stream=stream)
    contents = stream.getvalue()
    assert "Thread MainThread  (daemon: False, pydevd thread: False)" in contents
    assert "test_dump_threads" in contents



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import os
from collections.abc import MutableMapping, MutableSet


class DebugConfig(MutableMapping):
    """Debug configuration for a session. Corresponds to bodies of DAP "launch" and
    "attach" requests, or launch.json in VSCode.

    It is a dict-like object that only allows keys that are valid debug configuration
    properties for debugpy. When a property is queried, but it's not explicitly set in
    the config, the default value (i.e. what debugpy will assume the property is set to)
    is returned.

    In addition, it exposes high-level wrappers over "env" and "debugOptions".
    """

    __slots__ = ["_dict", "_env", "_debug_options"]

    # Valid configuration properties. Keys are names, and values are defaults that
    # are assumed by the adapter and/or the server if the property is not specified.
    # If the property is required, or if the default is computed in such a way that
    # it cannot be predicted, the value is ().
    PROPERTIES = {
        # Common
        "breakOnSystemExitZero": False,
        "debugOptions": [],
        "django": False,
        "jinja": False,
        "justMyCode": False,
        "flask": False,
        "logToFile": False,
        "maxExceptionStackFrames": (),
        "name": (),
        "noDebug": False,
        "pathMappings": [],
        "postDebugTask": (),
        "preLaunchTask": (),
        "pyramid": False,
        "redirectOutput": False,
        "rules": [],
        "showReturnValue": True,
        "steppingResumesAllThreads": True,
        "subProcess": False,
        "successExitCodes": [0],
        "type": (),
        # Launch
        "args": [],
        "code": (),
        "console": "internal",
        "cwd": (),
        "debugLauncherPython": (),
        "env": {},
        "gevent": False,
        "internalConsoleOptions": "neverOpen",
        "module": (),
        "program": (),
        "python": (),
        "pythonArgs": [],
        "pythonPath": (),
        "stopOnEntry": False,
        "sudo": False,
        "waitOnNormalExit": False,
        "waitOnAbnormalExit": False,
        # Attach by socket
        "listen": (),
        "connect": (),
        # Attach by PID
        "processId": (),
        "variablePresentation": {
            "special": "group",
            "function": "group",
            "class_": "group",
            "protected": "inline",
        },
        "autoReload": {
            "enable": False,
            "watch_dirs": [],
            "pollingInterval": 1,
            "exclude": [
                "**/.git/**",
                "**/__pycache__/**",
                "**/node_modules/**",
                "**/.metadata/**",
                "**/site-packages/**",
            ],
            "include": ["**/*.py", "**/*.pyw"],
        },
    }

    def __init__(self, *args, **kwargs):
        self._dict = dict(*args, **kwargs)
        self._env = self.Env(self)
        self._debug_options = self.DebugOptions(self)

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __contains__(self, key):
        return key in self._dict

    def __getitem__(self, key):
        try:
            return self._dict[key]
        except KeyError:
            try:
                value = self.PROPERTIES[key]
            except KeyError:
                pass
            else:
                if value != ():
                    return value
            raise

    def __delitem__(self, key):
        del self._dict[key]

    def __setitem__(self, key, value):
        assert key in self.PROPERTIES
        self._dict[key] = value

    def __repr__(self):
        return repr(dict(self))

    def __getstate__(self):
        return dict(self)

    def setdefault(self, key, default=None):
        if key not in self:
            self[key] = default
        return self[key]

    def setdefaults(self, defaults):
        """Like setdefault(), but sets multiple default values at once."""
        for k, v in defaults.items():
            self.setdefault(k, v)

    def normalize(self):
        """Normalizes the debug configuration by adding any derived properties, in
        the manner similar to what VSCode does to launch.json before submitting it
        to the adapter.
        """

        if self["showReturnValue"]:
            self.debug_options.add("ShowReturnValue")

        if self["redirectOutput"]:
            self.debug_options.add("RedirectOutput")

        if not self["justMyCode"]:
            self.debug_options.add("DebugStdLib")

        if self["django"]:
            self.debug_options.add("Django")

        if self["jinja"]:
            self.debug_options.add("Jinja")

        if self["flask"]:
            self.debug_options.add("Flask")

        if self["pyramid"]:
            self.debug_options.add("Pyramid")

        if self["subProcess"]:
            self.debug_options.add("Multiprocess")

        if self["breakOnSystemExitZero"]:
            self.debug_options.add("BreakOnSystemExitZero")

        if self["stopOnEntry"]:
            self.debug_options.add("StopOnEntry")

        if self["waitOnNormalExit"]:
            self.debug_options.add("WaitOnNormalExit")

        if self["waitOnAbnormalExit"]:
            self.debug_options.add("WaitOnAbnormalExit")

    @property
    def env(self):
        return self._env

    @property
    def debug_options(self):
        return self._debug_options

    class Env(MutableMapping):
        """Wraps config["env"], automatically creating and destroying it as needed."""

        def __init__(self, config):
            self.config = config

        def __iter__(self):
            return iter(self.config["env"])

        def __len__(self):
            return len(self.config["env"])

        def __getitem__(self, key):
            return self.config["env"][key]

        def __delitem__(self, key):
            env = self.config.get("env", {})
            del env[key]
            if not len(env):
                del self.config["env"]

        def __setitem__(self, key, value):
            self.config.setdefault("env", {})[key] = value

        def __getstate__(self):
            return dict(self)

        def prepend_to(self, key, entry):
            """Prepends a new entry to a PATH-style environment variable, creating
            it if it doesn't exist already.
            """
            try:
                tail = os.path.pathsep + self[key]
            except KeyError:
                tail = ""
            self[key] = entry + tail

    class DebugOptions(MutableSet):
        """Wraps config["debugOptions"], automatically creating and destroying it as
        needed, and providing set operations for it.
        """

        def __init__(self, config):
            self.config = config

        def __iter__(self):
            return iter(self.config["debugOptions"])

        def __len__(self):
            return len(self.config["env"])

        def __contains__(self, key):
            return key in self.config["debugOptions"]

        def add(self, key):
            opts = self.config.setdefault("debugOptions", [])
            if key not in opts:
                opts.append(key)

        def discard(self, key):
            opts = self.config.get("debugOptions", [])
            opts[:] = [x for x in opts if x != key]



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import inspect
import os
import py
import pytest
import sys
import threading
import types

from debugpy.common import log, timestamp
import tests
from tests import code, logs
from tests.debug import runners, session, targets

# Set up the test matrix for various code types and attach methods

if tests.full:
    TARGETS = targets.all_named
    RUNNERS = runners.all_launch + runners.all_attach_socket
else:
    TARGETS = [targets.Program]
    RUNNERS = [runners.launch]


@pytest.fixture(params=TARGETS)
def target(request):
    return request.param


@pytest.fixture(params=RUNNERS)
def run(request):
    return request.param


@pytest.fixture(autouse=True)
def test_wrapper(request, long_tmpdir):
    def write_log(filename, data):
        filename = os.path.join(log.log_dir, filename)
        if not isinstance(data, bytes):
            data = data.encode("utf-8")
        with open(filename, "wb") as f:
            f.write(data)

    session.Session.reset_counter()

    session.Session.tmpdir = long_tmpdir
    original_log_dir = log.log_dir

    failed = True
    try:
        if log.log_dir is None:
            log.log_dir = (long_tmpdir / "debugpy_logs").strpath
        else:
            log_subdir = request.node.nodeid
            log_subdir = log_subdir.replace("::", "/")
            for ch in r":?*|<>":
                log_subdir = log_subdir.replace(ch, f"&#{ord(ch)};")
            log.log_dir += "/" + log_subdir

        try:
            py.path.local(log.log_dir).remove()
        except Exception:
            pass

        print("\n")  # make sure on-screen logs start on a new line
        with log.to_file(prefix="tests"):
            timestamp.reset()
            log.info("{0} started.", request.node.nodeid)
            log.describe_environment("Environment:")
            try:
                yield
            finally:
                failed = False
                for report_attr in ("setup_report", "call_report", "teardown_report"):
                    try:
                        report = getattr(request.node, report_attr)
                    except AttributeError:
                        continue

                    failed |= report.failed
                    log.write_format(
                        "error" if report.failed else "info",
                        "pytest {0} phase for {1} {2}.",
                        report.when,
                        request.node.nodeid,
                        report.outcome,
                    )

                    write_log(report_attr + ".log", report.longreprtext)
                    write_log(report_attr + ".stdout.log", report.capstdout)
                    write_log(report_attr + ".stderr.log", report.capstderr)

                if failed:
                    write_log("FAILED.log", "")
                    logs.dump()

    finally:
        if not failed and not request.config.option.debugpy_log_passed:
            # try:
            #     py.path.local(log.log_dir).remove()
            # except Exception:
            #     pass
            pass
        log.log_dir = original_log_dir


@pytest.fixture
def daemon(request):
    """Provides a factory function for daemon threads. The returned thread is
    started immediately, and it must not be alive by the time the test returns.
    """

    daemons = []

    def factory(func, name_suffix=""):
        name = func.__name__ + name_suffix
        thread = threading.Thread(target=func, name=name)
        thread.daemon = True
        daemons.append(thread)
        thread.start()
        return thread

    yield factory

    try:
        failed = request.node.call_report.failed
    except AttributeError:
        pass
    else:
        if not failed:
            for thread in daemons:
                assert not thread.is_alive()


if sys.platform != "win32":

    @pytest.fixture
    def long_tmpdir(request, tmpdir):
        return tmpdir

else:
    import ctypes

    GetLongPathNameW = ctypes.windll.kernel32.GetLongPathNameW
    GetLongPathNameW.argtypes = [ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32]
    GetLongPathNameW.restype = ctypes.c_uint32

    @pytest.fixture
    def long_tmpdir(request, tmpdir):
        """Like tmpdir, but ensures that it's a long rather than short filename on Win32."""
        path = tmpdir.strpath
        buffer = ctypes.create_unicode_buffer(512)
        if GetLongPathNameW(path, buffer, len(buffer)):
            path = buffer.value
        return py.path.local(path)


@pytest.fixture
def pyfile(request, long_tmpdir):
    """A fixture providing a factory function that generates .py files.

    The returned factory takes a single function with an empty argument list,
    generates a temporary file that contains the code corresponding to the
    function body, and returns the full path to the generated file. Idiomatic
    use is as a decorator, e.g.:

        @pyfile
        def script_file():
            print('fizz')
            print('buzz')

    will produce a temporary file named script_file.py containing:

        print('fizz')
        print('buzz')

    and the variable script_file will contain the path to that file.

    In order for the factory to be able to extract the function body properly,
    function header ("def") must all be on a single line, with nothing after
    the colon but whitespace.

    Note that because the code is physically in a separate file when it runs,
    it cannot reuse top-level module imports - it must import all the modules
    that it uses locally. When linter complains, use #noqa.

    Returns a py.path.local instance that has the additional attribute "lines".
    After the source is writen to disk, tests.code.get_marked_line_numbers() is
    invoked on the resulting file to compute the value of that attribute.
    """

    def factory(source):
        assert isinstance(source, types.FunctionType)
        name = source.__name__
        source, _ = inspect.getsourcelines(source)

        # First, find the "def" line.
        def_lineno = 0
        for line in source:
            line = line.strip()
            if line.startswith("def") and line.endswith(":"):
                break
            def_lineno += 1
        else:
            raise ValueError("Failed to locate function header.")

        # Remove everything up to and including "def".
        source = source[def_lineno + 1 :]
        assert source

        # Now we need to adjust indentation. Compute how much the first line of
        # the body is indented by, then dedent all lines by that amount. Blank
        # lines don't matter indentation-wise, and might not be indented to begin
        # with, so just replace them with a simple newline.
        line = source[0]
        indent = len(line) - len(line.lstrip())
        source = [s[indent:] if s.strip() else "\n" for s in source]
        source = "".join(source)

        # Add a sleep at the end so that the program doesn't exit before we can handle all of the messages it sent
        source += "\nimport time\ntime.sleep(2)\n"

        # Write it to file.
        tmpfile = long_tmpdir / (name + ".py")
        tmpfile.strpath = tmpfile.strpath
        assert not tmpfile.check()
        tmpfile.write(source)

        tmpfile.lines = code.get_marked_line_numbers(tmpfile)
        return tmpfile

    return factory



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest

from tests import debug
from tests.debug import runners


@pytest.mark.parametrize("run", runners.all)
@pytest.mark.parametrize("redirect_output", ["", "redirect_output"])
def test_stdin_not_patched(pyfile, target, run, redirect_output):
    @pyfile
    def code_to_debug():
        import sys
        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        backchannel.send(sys.stdin is sys.__stdin__)

    with debug.Session() as session:
        session.config["redirectOutput"] = bool(redirect_output)

        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            pass

        is_original_stdin = backchannel.receive()
        assert is_original_stdin, "Expected sys.stdin and sys.__stdin__ to be the same."


@pytest.mark.parametrize("run", runners.all_launch_terminal + runners.all_attach)
@pytest.mark.parametrize("redirect_output", ["", "redirect_output"])
def test_input(pyfile, target, run, redirect_output):
    @pyfile
    def code_to_debug():
        import debuggee
        from debuggee import backchannel

        try:
            input = raw_input
        except NameError:
            pass

        debuggee.setup()
        backchannel.send(input())

    with debug.Session() as session:
        session.config["redirectOutput"] = bool(redirect_output)

        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            pass

        session.debuggee.stdin.write(b"ok\n")
        s = backchannel.receive()
        assert s == "ok"



import os
import sys
from _pydevd_bundle.pydevd_constants import PYDEVD_USE_SYS_MONITORING
port = int(sys.argv[1])
root_dirname = os.path.dirname(os.path.dirname(__file__))

if root_dirname not in sys.path:
    sys.path.append(root_dirname)

import pydevd

# Ensure that pydevd uses JSON protocol
from _pydevd_bundle import pydevd_constants
from _pydevd_bundle import pydevd_defaults
pydevd_defaults.PydevdCustomization.DEFAULT_PROTOCOL = pydevd_constants.HTTP_JSON_PROTOCOL

# Enable some defaults related to debugpy such as sending a single notification when
# threads pause and stopping on any exception.
pydevd_defaults.PydevdCustomization.DEBUG_MODE = 'debugpy-dap'

import tempfile
with tempfile.TemporaryDirectory('w') as tempdir:
    with open(os.path.join(tempdir, 'my_custom_module.py'), 'w') as stream:
        stream.write("print('Loaded my_custom_module')")

    pydevd_defaults.PydevdCustomization.PREIMPORT = '%s;my_custom_module' % (tempdir,)
    assert 'my_custom_module' not in sys.modules

    assert sys.gettrace() is None
    print('enable attach to port: %s' % (port,))
    pydevd._enable_attach(('127.0.0.1', port))

    assert pydevd.get_global_debugger() is not None
    # Set as a part of debugpy-dap
    assert pydevd.get_global_debugger().multi_threads_single_notification
    if PYDEVD_USE_SYS_MONITORING:
        assert sys.monitoring.get_tool(sys.monitoring.DEBUGGER_ID) == 'pydevd'
    else:
        assert sys.gettrace() is not None

    assert 'my_custom_module' in sys.modules

    a = 10  # Break 1
    print('wait for attach')
    pydevd._wait_for_attach()

    a = 20  # Break 2

    print('TEST SUCEEDED!')



import pytest
import sys


@pytest.mark.skipif(sys.platform != "win32", reason="Windows-only test.")
def test_pydevd_api_breakpoints(tmpdir):
    from _pydevd_bundle.pydevd_api import PyDevdAPI
    from pydevd import PyDB
    import pydevd_file_utils

    api = PyDevdAPI()

    py_db = PyDB(set_as_global=False)

    dira = tmpdir.join("DirA")
    dira.mkdir()

    f = dira.join("filE.py")
    f.write_text(
        """
a = 1
b = 2
c = 3
""",
        "utf-8",
    )
    filename = str(f)

    result = api.add_breakpoint(
        py_db,
        filename,
        breakpoint_type="python-line",
        breakpoint_id=0,
        line=1,
        condition=None,
        func_name="None",
        expression=None,
        suspend_policy="NONE",
        hit_condition="",
        is_logpoint=False,
    )
    assert not result.error_code

    result = api.add_breakpoint(
        py_db,
        filename,
        breakpoint_type="python-line",
        breakpoint_id=1,
        line=2,
        condition=None,
        func_name="None",
        expression=None,
        suspend_policy="NONE",
        hit_condition="",
        is_logpoint=False,
    )
    assert not result.error_code

    canonical_path = pydevd_file_utils.canonical_normalized_path(filename)

    assert len(py_db.breakpoints[canonical_path]) == 2
    assert len(py_db.file_to_id_to_line_breakpoint[canonical_path]) == 2

    filename_replaced = filename.replace("DirA", "dira")
    api.remove_all_breakpoints(py_db, filename_replaced)
    assert not py_db.breakpoints
    assert not py_db.file_to_id_to_line_breakpoint



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import inspect

import debugpy
from debugpy.common import log


def test_docstrings():
    for attr in debugpy.__all__:
        log.info("Checking docstring for debugpy.{0}", attr)
        member = getattr(debugpy, attr)

        doc = inspect.getdoc(member)
        assert doc is not None, f"debugpy.{member} doesn't have a docstring"
        for lineno, line in enumerate(doc.split("\n")):
            assert len(line) <= 72



import os
import sys
import platform

TEST_CYTHON = os.getenv("PYDEVD_USE_CYTHON", None) == "YES"
PYDEVD_TEST_VM = os.getenv("PYDEVD_TEST_VM", None)

IS_PY36_OR_GREATER = sys.version_info[0:2] >= (3, 6)
IS_PY311_OR_GREATER = sys.version_info[0:2] >= (3, 11)
IS_PY311 = sys.version_info[0:2] == (3, 11)
IS_PY312 = sys.version_info[0:2] == (3, 12)
IS_CPYTHON = platform.python_implementation() == "CPython"
IS_PYPY = platform.python_implementation() == "PyPy"

TODO_PYPY = IS_PYPY  # Code which needs to be fixed in pypy.

IS_PY36 = False
if sys.version_info[0] == 3 and sys.version_info[1] == 6:
    IS_PY36 = True

TEST_DJANGO = False
TEST_FLASK = False
TEST_CHERRYPY = False
TEST_GEVENT = False

try:
    import django

    TEST_DJANGO = True
except:
    pass

try:
    import flask

    TEST_FLASK = True
except:
    pass

try:
    import cherrypy

    TEST_CHERRYPY = True
except:
    pass

try:
    import gevent

    TEST_GEVENT = True
except:
    pass



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest

from tests import debug
from tests.debug import targets
from tests.patterns import some


@pytest.mark.parametrize("jmc", ["jmc", ""])
@pytest.mark.parametrize("target", targets.all_named)
def test_justmycode_frames(pyfile, target, run, jmc):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()

        import this  # @bp

        assert this

    with debug.Session() as session:
        session.config["justMyCode"] = bool(jmc)

        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        stop = session.wait_for_stop(
            "breakpoint", expected_frames=[some.dap.frame(code_to_debug, "bp")]
        )
        if jmc:
            assert len(stop.frames) == 1
        else:
            assert len(stop.frames) >= 1

        session.request("stepIn", {"threadId": stop.thread_id})

        # With JMC, it should step out of the function, remaining in the same file.
        # Without JMC, it should step into stdlib.
        expected_path = some.path(code_to_debug)
        if not jmc:
            expected_path = ~expected_path
        session.wait_for_stop(
            "step",
            expected_frames=[some.dap.frame(some.dap.source(expected_path), some.int)],
        )

        session.request_continue()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""Runners are recipes for executing Targets in a debug.Session.

Every function in this module that is decorated with @_runner must have at least two
positional arguments: (session, target) - and can have additional arguments. For every
such function, two artifacts are produced.

The function is exposed directly as a method on Session, with the session argument
becoming self.

The function is also exposed as a Runner object from this module. Runner objects are
callable, and invoke the wrapped function when called, but in addition, they can also
be bound to specific arguments, by using either [] or with_options(), which can be
chained arbitrarily::

    # Direct invocation:
    session.attach_connect("cli", log_dir="...")

    # Indirect invocation:
    run = runners.attach_connect
    run = run["cli"]
    run = run.with_options(log_dir="...")
    run(session, target)

runner[x][y][z] is just a convenient shorthand for binding positional arguments, same
as runner.with_options(x, y, z).

Runners are immutable, so every use of [] or with_options() creates a new runner with
the specified arguments bound. The runner must have all its required arguments bound
before it can be invoked.

Regardless of whether the runner is invoked directly on the Session, or via a Runner
object, if the start DAP sequence involves a configuration phase (the "initialized"
event and the "configurationDone" request), the runner must be used in a with-statement.
The statements inside the with-statement are executed after receiving the "initialized"
event, and before sending the "configurationDone" request::

    with run(session, target):
        # DAP requests can be made to session, but target is not running yet.
        session.set_breakpoints(...)
    # target is running now!

If there is no configuration phase, the runner returns directly::

    session.config["noDebug"] = True
    run(session, target)
    # target is running now!
"""

import contextlib
import os
import pytest
import sys

import debugpy
from debugpy.common import json, log
from tests import net, timeline
from tests.debug import session
from tests.patterns import some


def _runner(f):
    # assert f.__name__.startswith("launch") or f.__name__.startswith("attach")
    setattr(session.Session, f.__name__, f)

    class Runner(object):
        request = "launch" if f.__name__.startswith("launch") else "attach"

        def __init__(self, *args, **kwargs):
            self._args = tuple(args)
            self._kwargs = dict(kwargs)

        def __getattr__(self, name):
            return self._kwargs[name]

        def __call__(self, session, target, *args, **kwargs):
            if len(args) or len(kwargs):
                return self.with_options(*args, **kwargs)(session, target)
            return f(session, target, *self._args, **self._kwargs)

        def __iter__(self):
            # Since we implement __getitem__, iter() will assume that runners are
            # iterable, and will iterate over them by calling __getitem__ until it
            # raises IndexError - i.e. indefinitely. To prevent that, explicitly
            # implement __iter__ as unsupported.
            raise NotImplementedError

        def __getitem__(self, arg):
            return self.with_options(arg)

        def with_options(self, *args, **kwargs):
            new_args = self._args + args
            new_kwargs = dict(self._kwargs)
            new_kwargs.update(kwargs)
            return Runner(*new_args, **new_kwargs)

        def __repr__(self):
            result = type(self).__name__
            args = [str(x) for x in self._args] + [
                f"{k}={v}" for k, v in self._kwargs.items()
            ]
            if len(args):
                result += "(" + ", ".join(args) + ")"
            return result

        @property
        def pytest_id(self):
            return repr(self)

    Runner.__name__ = f.__name__
    return Runner()


@_runner
def launch(session, target, console=None, cwd=None):
    assert console in (
        None,
        "internalConsole",
        "integratedTerminal",
        "externalTerminal",
    )

    log.info("Launching {0} in {1} using {2}.", target, session, json.repr(console))

    target.configure(session)
    config = session.config
    config.setdefaults(
        {"console": "externalTerminal", "internalConsoleOptions": "neverOpen"}
    )
    if console is not None:
        config["console"] = console
    if cwd is not None:
        config["cwd"] = cwd
    if "python" not in config and "pythonPath" not in config:
        config["python"] = sys.executable

    env = (
        session.spawn_adapter.env
        if config["console"] == "internalConsole"
        else config.env
    )
    target.cli(env)

    session.spawn_adapter()
    return session.request_launch()


def _attach_common_config(session, target, cwd):
    assert (
        target.code is None or "debuggee.setup()" in target.code
    ), f"{target.filename} must invoke debuggee.setup()."

    target.configure(session)
    config = session.config
    if cwd is not None:
        config.setdefault("pathMappings", [{"localRoot": cwd, "remoteRoot": "."}])
    return config


@_runner
@contextlib.contextmanager
def attach_pid(session, target, cwd=None, wait=True):
    if wait and not sys.platform.startswith("linux"):
        pytest.skip("https://github.com/microsoft/ptvsd/issues/1926")

    log.info("Attaching {0} to {1} by PID.", session, target)

    config = session.config
    try:
        config["processId"] = int(target)
    except TypeError:
        pass

    if "processId" not in config:
        _attach_common_config(session, target, cwd)
        args = target.cli(session.spawn_debuggee.env)

        if wait:
            debuggee_setup = """
import sys
import threading
import time

while "debugpy" not in sys.modules:
    time.sleep(0.1)

from debuggee import scratchpad

while "_attach_pid" not in scratchpad:
    time.sleep(0.1)
    """
        else:
            debuggee_setup = None

        session.spawn_debuggee(args, cwd=cwd, setup=debuggee_setup)
        config["processId"] = session.debuggee.pid

    session.spawn_adapter()
    session.expect_server_socket()
    with session.request_attach():
        yield

    if wait:
        session.scratchpad["_attach_pid"] = True


@_runner
def attach_connect(session, target, method, cwd=None, wait=True, log_dir=None):
    log.info(
        "Attaching {0} to {1} by socket using {2}.", session, target, method.upper()
    )

    assert method in ("api", "cli")

    config = _attach_common_config(session, target, cwd)
    config["connect"] = {}
    config["connect"]["host"] = host = attach_connect.host
    config["connect"]["port"] = port = attach_connect.port

    if method == "cli":
        args = [
            os.path.dirname(debugpy.__file__),
            "--listen",
            f"{host}:{port}",
        ]
        if wait:
            args += ["--wait-for-client"]
        if log_dir is not None:
            args += ["--log-to", log_dir]
        if "subProcess" in config:
            args += ["--configure-subProcess", str(config["subProcess"])]
        debuggee_setup = None
    elif method == "api":
        args = []
        api_config = {k: v for k, v in config.items() if k in {"subProcess"}}
        debuggee_setup = """
import debugpy
if {log_dir!r}:
    debugpy.log_to({log_dir!r})
debugpy.configure({api_config!r})
debugpy.listen(({host!r}, {port!r}))
if {wait!r}:
    debugpy.wait_for_client()
"""
        debuggee_setup = debuggee_setup.format(
            host=host,
            port=port,
            wait=wait,
            log_dir=log_dir,
            api_config=api_config,
        )
    else:
        raise ValueError
    args += target.cli(session.spawn_debuggee.env)

    try:
        del config["subProcess"]
    except KeyError:
        pass

    # If adapter is connecting to the client, the server is already started,
    # so it should be reported in the initial event.
    session.expect_server_socket()

    session.spawn_debuggee(args, cwd=cwd, setup=debuggee_setup)
    session.wait_for_adapter_socket()
    session.connect_to_adapter((host, port))
    return session.request_attach()


attach_connect.host = "127.0.0.1"
attach_connect.port = net.get_test_server_port()


@_runner
def attach_listen(session, target, method, cwd=None, log_dir=None):
    log.info(
        "Attaching {0} to {1} by socket using {2}.", session, target, method.upper()
    )

    assert method in ("api", "cli")

    config = _attach_common_config(session, target, cwd)
    config["listen"] = {}
    config["listen"]["host"] = host = attach_listen.host
    config["listen"]["port"] = port = attach_listen.port

    if method == "cli":
        args = [
            os.path.dirname(debugpy.__file__),
            "--connect",
            f"{host}:{port}",
        ]
        if log_dir is not None:
            args += ["--log-to", log_dir]
        if "subProcess" in config:
            args += ["--configure-subProcess", str(config["subProcess"])]
        debuggee_setup = None
    elif method == "api":
        args = []
        api_config = {k: v for k, v in config.items() if k in {"subProcess"}}
        debuggee_setup = f"""
import debugpy
if {log_dir!r}:
    debugpy.log_to({log_dir!r})
debugpy.configure({api_config!r})
debugpy.connect({(host, port)!r})
"""
    else:
        raise ValueError
    args += target.cli(session.spawn_debuggee.env)

    try:
        del config["subProcess"]
    except KeyError:
        pass

    def spawn_debuggee(occ):
        assert occ.body == some.dict.containing({"host": host, "port": port})
        session.spawn_debuggee(args, cwd=cwd, setup=debuggee_setup)

    session.timeline.when(timeline.Event("debugpyWaitingForServer"), spawn_debuggee)
    session.spawn_adapter(args=[] if log_dir is None else ["--log-dir", log_dir])
    return session.request_attach()


attach_listen.host = "127.0.0.1"
attach_listen.port = net.get_test_server_port()

all_launch_terminal = [
    launch.with_options(console="integratedTerminal"),
    launch.with_options(console="externalTerminal"),
]

all_launch = [launch.with_options(console="internalConsole")] + all_launch_terminal

all_attach_listen = [attach_listen["api"], attach_listen["cli"]]

all_attach_connect = [attach_connect["api"], attach_connect["cli"]]

all_attach_socket = all_attach_listen + all_attach_connect

all_attach = all_attach_socket + [attach_pid]

all = all_launch + all_attach



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest

from tests import debug, timeline
from tests.patterns import some


def test_evaluate(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()
        a = 1
        b = {"one": 1, 2: "two"}
        print(a, b)  # @bp

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        stop = session.wait_for_stop()

        # Evaluate a variable name.
        evaluate1 = session.request(
            "evaluate", {"expression": "a", "frameId": stop.frame_id}
        )
        assert evaluate1 == some.dict.containing({"type": "int", "result": "1"})

        # Evaluate dict indexing.
        evaluate2 = session.request(
            "evaluate", {"expression": "b[2]", "frameId": stop.frame_id}
        )
        assert evaluate2 == some.dict.containing({"type": "str", "result": "'two'"})

        # Evaluate an expression with a binary operator.
        evaluate3 = session.request(
            "evaluate", {"expression": 'a + b["one"]', "frameId": stop.frame_id}
        )
        assert evaluate3 == some.dict.containing({"type": "int", "result": "2"})

        session.request_continue()


def test_variables(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()
        a = 1
        b = {"one": 1, 2: "two"}
        c = 3
        print([a, b, c])  # @bp

    with debug.Session() as session:
        session.config["variablePresentation"] = {"all": "hide", "protected": "inline"}
        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        stop = session.wait_for_stop()
        scopes = session.request("scopes", {"frameId": stop.frame_id})["scopes"]
        globals_ref = scopes[0]["variablesReference"]
        vars = session.request("variables", {"variablesReference": globals_ref})[
            "variables"
        ]

        # Variables must be sorted by name.
        a, b, c = (v for v in vars if v["name"] in ("a", "b", "c"))
        assert (a["name"], b["name"], c["name"]) == ("a", "b", "c")

        # Fetch children variables of the dict.
        b_vars = session.request(
            "variables", {"variablesReference": b["variablesReference"]}
        )["variables"]
        assert b_vars == [
            some.dict.containing(
                {
                    "type": "int",
                    "value": "1",
                    "name": "'one'",
                    "evaluateName": "b['one']",
                    "variablesReference": 0,
                }
            ),
            some.dict.containing(
                {
                    "type": "str",
                    "value": "'two'",
                    "name": "2",
                    "evaluateName": "b[2]",
                    "variablesReference": 0,
                }
            ),
            some.dict.containing(
                {
                    "type": "int",
                    "value": "2",
                    "name": "len()",
                    "evaluateName": "len(b)",
                    "variablesReference": 0,
                    "presentationHint": {"attributes": ["readOnly"]},
                }
            ),
        ]

        session.request_continue()


def test_variable_sort(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()
        b_test = {"spam": "A", "eggs": "B", "abcd": "C"}  # noqa
        _b_test = 12  # noqa
        __b_test = 13  # noqa
        __b_test__ = 14  # noqa
        a_test = 1  # noqa
        _a_test = 2  # noqa
        __a_test = 3  # noqa
        __a_test__ = 4  # noqa
        c_test = {1: "one", 2: "two", 10: "ten"}  # noqa
        _c_test = 22  # noqa
        __c_test = 23  # noqa
        __c_test__ = 24  # noqa
        d = 3  # noqa
        print("done")  # @bp

    with debug.Session() as session:
        session.config["variablePresentation"] = {
            "special": "group",
            "protected": "inline",
        }
        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        stop = session.wait_for_stop()
        scopes = session.request("scopes", {"frameId": stop.frame_id})["scopes"]
        globals_ref = scopes[0]["variablesReference"]
        vars = session.request("variables", {"variablesReference": globals_ref})[
            "variables"
        ]

        var_names = [v["name"] for v in vars if "_test" in v["name"]]
        assert var_names == [
            "a_test",
            "b_test",
            "c_test",
            "_a_test",
            "_b_test",
            "_c_test",
            "__a_test",
            "__b_test",
            "__c_test",
        ]

        (special_vars_entry,) = (v for v in vars if v["name"] == "special variables")
        special_vars_variables = session.request(
            "variables",
            {"variablesReference": special_vars_entry["variablesReference"]},
        )["variables"]
        special_vars_variables = [
            v["name"] for v in special_vars_variables if "_test" in v["name"]
        ]
        assert special_vars_variables == ["__a_test__", "__b_test__", "__c_test__"]

        # String dict keys must be sorted as strings.
        (b_test,) = (v for v in vars if v["name"] == "b_test")
        b_test_vars = session.request(
            "variables", {"variablesReference": b_test["variablesReference"]}
        )["variables"]
        var_names = [v["name"] for v in b_test_vars]
        # Note that the special len() we manually create is not added to special variables.
        expected = [
            "special variables",
            "function variables",
            "'spam'",
            "'eggs'",
            "'abcd'",
            "len()",
        ]
        assert var_names == expected

        # Numeric dict keys must be sorted as numbers.
        if not "https://github.com/microsoft/ptvsd/issues/213":
            (c_test,) = (v for v in vars if v["name"] == "c_test")
            c_test_vars = session.request(
                "variables", {"variablesReference": c_test["variablesReference"]}
            )["variables"]
            var_names = [v["name"] for v in c_test_vars]
            # Note that the special len() we manually create is not added to special variables.
            expected = ["1", "2", "10", "len()"]

            assert var_names == expected

        session.request_continue()


@pytest.mark.parametrize("ret_vis", ("show", "hide", "default"))
def test_return_values(pyfile, target, run, ret_vis):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()

        class MyClass(object):
            def do_something(self):
                return "did something"

        def my_func():
            return "did more things"

        MyClass().do_something()  # @bp
        my_func()
        print("done")

    expected1 = some.dict.containing(
        {
            "name": "(return) MyClass.do_something",
            "value": "'did something'",
            "type": "str",
            "presentationHint": some.dict.containing(
                {"attributes": some.list.containing("readOnly")}
            ),
        }
    )

    expected2 = some.dict.containing(
        {
            "name": "(return) my_func",
            "value": "'did more things'",
            "type": "str",
            "presentationHint": some.dict.containing(
                {"attributes": some.list.containing("readOnly")}
            ),
        }
    )

    with debug.Session() as session:
        if ret_vis != "default":
            session.config["showReturnValue"] = ret_vis != "hide"

        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        stop = session.wait_for_stop()
        session.request("next", {"threadId": stop.thread_id})
        stop = session.wait_for_stop("step")
        scopes = session.request("scopes", {"frameId": stop.frame_id})["scopes"]
        globals_ref = scopes[0]["variablesReference"]

        vars = session.request("variables", {"variablesReference": globals_ref})[
            "variables"
        ]
        ret_vars = [v for v in vars if v["name"].startswith("(return)")]
        assert ret_vars == ([expected1] if ret_vis != "hide" else [])

        session.request("next", {"threadId": stop.thread_id})
        stop = session.wait_for_stop("step")

        # Variable reference for the scope is not invalidated after the step.
        vars = session.request("variables", {"variablesReference": globals_ref})[
            "variables"
        ]
        ret_vars = [v for v in vars if v["name"].startswith("(return)")]
        assert ret_vars == ([expected1, expected2] if ret_vis != "hide" else [])

        session.request_continue()


# On Python 3, variable names can contain Unicode characters.
def test_unicode(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee
        import debugpy

        globals()["\u16A0"] = 123
        debuggee.setup()
        debugpy.breakpoint()
        print("break")

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            pass

        stop = session.wait_for_stop()
        eval = session.request(
            "evaluate", {"expression": "\u16A0", "frameId": stop.frame_id}
        )
        assert eval == some.dict.containing({"type": "int", "result": "123"})
        session.request_continue()


# Numbers should be properly hex-formatted in all positions: variable values, list
# indices, dict keys etc.
def test_hex_numbers(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()
        a = 100
        b = [1, 10, 100]
        c = {10: 10, 100: 100, 1000: 1000}
        d = {(1, 10, 100): (10000, 100000, 100000)}
        print((a, b, c, d))  # @bp

    with debug.Session() as session:
        session.config["variablePresentation"] = {"all": "hide", "protected": "inline"}

        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        stop = session.wait_for_stop()
        scopes = session.request("scopes", {"frameId": stop.frame_id})["scopes"]
        globals_ref = scopes[0]["variablesReference"]

        vars = session.request(
            "variables", {"variablesReference": globals_ref, "format": {"hex": True}}
        )["variables"]
        a, b, c, d = (v for v in vars if v["name"] in ("a", "b", "c", "d"))
        assert a == some.dict.containing(
            {
                "name": "a",
                "value": "0x64",
                "type": "int",
                "evaluateName": "a",
                "variablesReference": 0,
            }
        )
        assert b == some.dict.containing(
            {
                "name": "b",
                "value": "[0x1, 0xa, 0x64]",
                "type": "list",
                "evaluateName": "b",
                "variablesReference": some.dap.id,
            }
        )
        assert c == some.dict.containing(
            {
                "name": "c",
                "value": "{0xa: 0xa, 0x64: 0x64, 0x3e8: 0x3e8}",
                "type": "dict",
                "evaluateName": "c",
                "variablesReference": some.dap.id,
            }
        )
        assert d == some.dict.containing(
            {
                "name": "d",
                "value": "{(0x1, 0xa, 0x64): (0x2710, 0x186a0, 0x186a0)}",
                "type": "dict",
                "evaluateName": "d",
                "variablesReference": some.dap.id,
            }
        )

        b_vars = session.request(
            "variables",
            {"variablesReference": b["variablesReference"], "format": {"hex": True}},
        )["variables"]
        assert b_vars == [
            some.dict.containing(
                {
                    "name": "0x0",
                    "value": "0x1",
                    "type": "int",
                    "evaluateName": "b[0]",
                    "variablesReference": 0,
                }
            ),
            some.dict.containing(
                {
                    "name": "0x1",
                    "value": "0xa",
                    "type": "int",
                    "evaluateName": "b[1]",
                    "variablesReference": 0,
                }
            ),
            some.dict.containing(
                {
                    "name": "0x2",
                    "value": "0x64",
                    "type": "int",
                    "evaluateName": "b[2]",
                    "variablesReference": 0,
                }
            ),
            some.dict.containing(
                {
                    "name": "len()",
                    "value": "0x3",
                    "type": "int",
                    "evaluateName": "len(b)",
                    "variablesReference": 0,
                    "presentationHint": {"attributes": ["readOnly"]},
                }
            ),
        ]

        c_vars = session.request(
            "variables",
            {"variablesReference": c["variablesReference"], "format": {"hex": True}},
        )["variables"]
        # Use dict sequence on Python 3.6 onwards.
        assert c_vars == [
            some.dict.containing(
                {
                    "name": "0xa",
                    "value": "0xa",
                    "type": "int",
                    "evaluateName": "c[10]",
                    "variablesReference": 0,
                }
            ),
            some.dict.containing(
                {
                    "name": "0x64",
                    "value": "0x64",
                    "type": "int",
                    "evaluateName": "c[100]",
                    "variablesReference": 0,
                }
            ),
            some.dict.containing(
                {
                    "name": "0x3e8",
                    "value": "0x3e8",
                    "type": "int",
                    "evaluateName": "c[1000]",
                    "variablesReference": 0,
                }
            ),
            some.dict.containing(
                {
                    "name": "len()",
                    "value": "0x3",
                    "type": "int",
                    "evaluateName": "len(c)",
                    "variablesReference": 0,
                    "presentationHint": {"attributes": ["readOnly"]},
                }
            ),
        ]

        d_vars = session.request(
            "variables",
            {"variablesReference": d["variablesReference"], "format": {"hex": True}},
        )["variables"]
        assert d_vars == [
            some.dict.containing(
                {
                    "name": "(0x1, 0xa, 0x64)",
                    "value": "(0x2710, 0x186a0, 0x186a0)",
                    "type": "tuple",
                    "evaluateName": "d[(1, 10, 100)]",
                    "variablesReference": some.dap.id,
                }
            ),
            some.dict.containing(
                {
                    "name": "len()",
                    "value": "0x1",
                    "type": "int",
                    "evaluateName": "len(d)",
                    "variablesReference": 0,
                    "presentationHint": {"attributes": ["readOnly"]},
                }
            ),
        ]

        d_item0 = d_vars[0]
        d_item0_vars = session.request(
            "variables",
            {
                "variablesReference": d_item0["variablesReference"],
                "format": {"hex": True},
            },
        )["variables"]
        assert d_item0_vars == [
            some.dict.containing(
                {
                    "name": "0x0",
                    "value": "0x2710",
                    "type": "int",
                    "evaluateName": "d[(1, 10, 100)][0]",
                    "variablesReference": 0,
                }
            ),
            some.dict.containing(
                {
                    "name": "0x1",
                    "value": "0x186a0",
                    "type": "int",
                    "evaluateName": "d[(1, 10, 100)][1]",
                    "variablesReference": 0,
                }
            ),
            some.dict.containing(
                {
                    "name": "0x2",
                    "value": "0x186a0",
                    "type": "int",
                    "evaluateName": "d[(1, 10, 100)][2]",
                    "variablesReference": 0,
                }
            ),
            some.dict.containing(
                {
                    "name": "len()",
                    "value": "0x3",
                    "type": "int",
                    "evaluateName": "len(d[(1, 10, 100)])",
                    "variablesReference": 0,
                    "presentationHint": {"attributes": ["readOnly"]},
                }
            ),
        ]

        session.request_continue()


def test_set_variable(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee
        import debugpy
        from debuggee import backchannel

        debuggee.setup()
        a = 1
        debugpy.breakpoint()
        backchannel.send(a)

    with debug.Session() as session:
        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            pass

        stop = session.wait_for_stop()
        scopes = session.request("scopes", {"frameId": stop.frame_id})["scopes"]
        globals_ref = scopes[0]["variablesReference"]
        vars = session.request("variables", {"variablesReference": globals_ref})[
            "variables"
        ]

        (a,) = (v for v in vars if v["name"] == "a")
        assert a == some.dict.containing(
            {
                "type": "int",
                "value": "1",
                "name": "a",
                "evaluateName": "a",
                "variablesReference": 0,
            }
        )

        set_a = session.request(
            "setVariable",
            {"variablesReference": globals_ref, "name": "a", "value": "1000"},
        )
        assert set_a == some.dict.containing({"type": "int", "value": "1000"})

        session.request_continue()
        assert backchannel.receive() == 1000


def test_set_expression(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        a = 1
        backchannel.send(a)  # @bp

    with debug.Session() as session:
        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        stop = session.wait_for_stop()
        scopes = session.request("scopes", {"frameId": stop.frame_id})["scopes"]
        globals_ref = scopes[0]["variablesReference"]

        vars = session.request("variables", {"variablesReference": globals_ref})[
            "variables"
        ]
        (a,) = (v for v in vars if v["name"] == "a")
        assert a == some.dict.containing(
            {
                "type": "int",
                "value": "1",
                "name": "a",
                "evaluateName": "a",
                "variablesReference": 0,
            }
        )

        set_a = session.request(
            "setExpression",
            {"frameId": stop.frame_id, "expression": "a", "value": "1000"},
        )
        assert set_a == some.dict.containing({"type": "int", "value": "1000"})

        session.request_continue()
        assert backchannel.receive() == 1000


def test_evaluate_thread_locks(pyfile, target, run):
    @pyfile
    def code_to_debug():
        """
        The idea here is that a secondary thread does the processing of instructions,
        so, when all threads are stopped, doing an evaluation for:

        processor.process('xxx')

        would be locked until secondary threads start running.
        See: https://github.com/microsoft/debugpy/issues/157
        """

        import debuggee
        import queue
        import threading

        debuggee.setup()

        class EchoThread(threading.Thread):
            def __init__(self, queue):
                threading.Thread.__init__(self)
                self._queue = queue

            def run(self):
                while True:
                    obj = self._queue.get()
                    if obj == "finish":
                        break

                    print("processed", obj.value)
                    obj.event.set()

        class NotificationObject(object):
            def __init__(self, value):
                self.value = value
                self.event = threading.Event()

        class Processor(object):
            def __init__(self, queue):
                self._queue = queue

            def process(self, i):
                obj = NotificationObject(i)
                self._queue.put(obj)
                assert obj.event.wait()

            def finish(self):
                self._queue.put("finish")

        if __name__ == "__main__":
            q = queue.Queue()
            echo_thread = EchoThread(q)
            processor = Processor(q)
            echo_thread.start()

            processor.process(1)
            processor.process(2)  # @bp
            processor.process(3)
            processor.finish()

    with debug.Session() as session:

        # During the evaluation we'll actually have continued/stopped events because
        # we're letting threads run at that time. Let's ignore these in the test.
        session.ignore_unobserved.extend([timeline.Event("stopped")])

        session.config.env.update({"PYDEVD_UNBLOCK_THREADS_TIMEOUT": "0.5"})

        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        stop = session.wait_for_stop()

        evaluate = session.request(
            "evaluate",
            {"expression": "processor.process('foo')", "frameId": stop.frame_id},
        )
        assert evaluate == some.dict.containing({"result": "None"})

        session.request_continue()



import sys
import pytest
import threading

pytestmark = pytest.mark.skipif(not hasattr(sys, "monitoring"), reason="Requires sys.monitoring")

if hasattr(sys, "monitoring"):
    DEBUGGER_ID = sys.monitoring.DEBUGGER_ID
    monitor = sys.monitoring


def _disable_monitoring():
    if monitor.get_tool(DEBUGGER_ID) == "pydevd":
        sys.monitoring.set_events(sys.monitoring.DEBUGGER_ID, 0)
        monitor.register_callback(DEBUGGER_ID, monitor.events.PY_START, None)
        monitor.register_callback(DEBUGGER_ID, monitor.events.PY_RESUME, None)
        monitor.register_callback(DEBUGGER_ID, monitor.events.LINE, None)
        monitor.register_callback(DEBUGGER_ID, monitor.events.RAISE, None)
        monitor.register_callback(DEBUGGER_ID, monitor.events.PY_RETURN, None)
        monitor.register_callback(DEBUGGER_ID, monitor.events.PY_UNWIND, None)
        sys.monitoring.free_tool_id(DEBUGGER_ID)


@pytest.fixture
def with_monitoring():
    monitor.use_tool_id(DEBUGGER_ID, "pydevd")
    yield
    _disable_monitoring()


def test_exceptions(with_monitoring):
    monitor.set_events(DEBUGGER_ID, monitor.events.RAISE | monitor.events.RERAISE)

    found = []

    def _on_raise(code, instruction_offset, exc):
        if code.co_filename.endswith("sys_monitoring.py"):
            found.append(("raise", code.co_name, str(exc), sys._getframe(1).f_lineno))

    def _on_reraise(code, instruction_offset, exc):
        if code.co_filename.endswith("sys_monitoring.py"):
            found.append(("reraise", code.co_name, str(exc), sys._getframe(1).f_lineno))

    monitor.register_callback(DEBUGGER_ID, monitor.events.RAISE, _on_raise)
    monitor.register_callback(DEBUGGER_ID, monitor.events.RERAISE, _on_reraise)

    def method_raise():
        raise RuntimeError("err1")

    def method_2():
        try:
            method_raise()
        except:
            raise

    def method():
        try:
            method_2()
        except:
            pass

    method()
    assert found == [
        ("raise", "method_raise", "err1", method_raise.__code__.co_firstlineno + 1),
        ("raise", "method_2", "err1", method_2.__code__.co_firstlineno + 2),
        # This will be very tricky to handle.
        # See: https://github.com/python/cpython/issues/112086
        ("reraise", "method_2", "err1", method_2.__code__.co_firstlineno + 4),
        ("reraise", "method_2", "err1", method_2.__code__.co_firstlineno + 4),
        ("raise", "method", "err1", method.__code__.co_firstlineno + 2),
    ]


def test_exceptions_and_return(with_monitoring):
    monitor.set_events(DEBUGGER_ID, monitor.events.RAISE | monitor.events.PY_RETURN | monitor.events.PY_UNWIND)

    found = []

    def _on_raise(code, instruction_offset, exc):
        if code.co_filename.endswith("sys_monitoring.py"):
            found.append(("raise", code.co_name, str(exc), sys._getframe(1).f_lineno))

    def _on_return(code, instruction_offset, val):
        if code.co_filename.endswith("sys_monitoring.py"):
            found.append(("return", code.co_name, str(val), sys._getframe(1).f_lineno))

    def _on_unwind(code, instruction_offset, val):
        if code.co_filename.endswith("sys_monitoring.py"):
            found.append(("unwind", code.co_name, str(val), sys._getframe(1).f_lineno))

    monitor.register_callback(DEBUGGER_ID, monitor.events.RAISE, _on_raise)
    monitor.register_callback(DEBUGGER_ID, monitor.events.PY_RETURN, _on_return)
    monitor.register_callback(DEBUGGER_ID, monitor.events.PY_UNWIND, _on_unwind)

    def method_raise():
        raise RuntimeError("err1")

    def method_2():
        try:
            method_raise()
        except:
            raise

    def method():
        try:
            method_2()
        except:
            pass

    method()

    assert found == [
        ("raise", "method_raise", "err1", 96),
        ("unwind", "method_raise", "err1", 96),
        ("raise", "method_2", "err1", 100),
        ("unwind", "method_2", "err1", 102),  # This will be helpful for unhandled exceptions!
        ("raise", "method", "err1", 106),
        ("return", "method", "None", 108),
    ]


def test_variables_on_call(with_monitoring):
    monitor.set_events(DEBUGGER_ID, monitor.events.PY_START)

    found = []

    def _start_method(code, offset):
        if code.co_name == "method":
            frame = sys._getframe(1)
            found.append(frame.f_locals["arg1"])

    monitor.register_callback(DEBUGGER_ID, monitor.events.PY_START, _start_method)

    def method(arg1):
        pass

    method(22)
    assert found == [22]


def test_disabling_code(with_monitoring):
    executed = []

    def _start_method(code, offset):
        if code.co_name == "method":
            executed.append(("start", code.co_name, offset))
            monitor.set_local_events(DEBUGGER_ID, code, monitor.events.LINE)
        return monitor.DISABLE

    def _on_line(code, offset):
        if code.co_name == "method":
            executed.append(("line", code.co_name, offset))
        return monitor.DISABLE

    monitor.set_events(DEBUGGER_ID, monitor.events.PY_START | monitor.events.PY_RESUME)

    monitor.register_callback(DEBUGGER_ID, monitor.events.PY_START, _start_method)
    monitor.register_callback(DEBUGGER_ID, monitor.events.PY_RESUME, _start_method)
    monitor.register_callback(DEBUGGER_ID, monitor.events.LINE, _on_line)

    def method():
        a = 1

    method()
    method()
    assert executed == [("start", "method", 0), ("line", "method", method.__code__.co_firstlineno + 1)]

    del executed[:]

    # Check: if disable once, even on a new thread we won't get notifications!
    t = threading.Thread(target=method)
    t.start()
    t.join()

    assert not executed

    # Unless restart_events is called...
    monitor.restart_events()
    t = threading.Thread(target=method)
    t.start()
    t.join()
    assert executed == [("start", "method", 0), ("line", "method", method.__code__.co_firstlineno + 1)]


def test_change_line_during_trace(with_monitoring):
    code_to_break_at_line = {}
    do_change_line = [0]
    lines_traced = []

    def _start_method(code, offset):
        monitor.set_local_events(DEBUGGER_ID, code, monitor.events.LINE)
        code_to_break_at_line[code] = {code.co_firstlineno + 3}
        return monitor.DISABLE

    def _on_line(code, line):
        lines_to_break = code_to_break_at_line.get(code)
        if lines_to_break and line in lines_to_break:
            do_change_line[0] += 1
            if do_change_line[0] == 2:
                frame = sys._getframe(1)
                frame.f_lineno = line - 2

    monitor.set_events(DEBUGGER_ID, monitor.events.PY_START | monitor.events.PY_RESUME)

    monitor.register_callback(DEBUGGER_ID, monitor.events.PY_START, _start_method)
    monitor.register_callback(DEBUGGER_ID, monitor.events.PY_RESUME, _start_method)
    monitor.register_callback(DEBUGGER_ID, monitor.events.LINE, _on_line)

    def method1():  # code.co_firstlineno
        a = 1  # code.co_firstlineno + 1
        lines_traced.append("before a=2")  # code.co_firstlineno + 2
        a = 2  # code.co_firstlineno + 3
        lines_traced.append("before a=3")  # code.co_firstlineno + 4
        # a = 3  # code.co_firstlineno + 5

    for _i in range(3):
        method1()

    assert lines_traced == [
        "before a=2",
        "before a=3",
        "before a=2",
        "before a=2",
        "before a=3",
        "before a=2",
        "before a=3",
    ]


def test_tracing():
    import sys

    def method():
        a = [1, 2, 3, 4, 5, 6]  # line 1

        # line 3
        def b():  #  line 4
            yield from [j for j in a if j % 2 == 0]  # line 5

        # line 7
        for j in b():  # line 8
            print(j)  # line 9

    def tracefunc(frame, event, arg):
        if "test_sys_monitoring.py" in frame.f_code.co_filename:
            print(frame.f_code.co_name, event, frame.f_lineno - frame.f_code.co_firstlineno)
            return tracefunc

    sys.settrace(tracefunc)
    method()
    sys.settrace(None)


def test_lines_with_yield(with_monitoring):
    def _start_method(code, offset):
        if "test_sys_monitoring.py" in code.co_filename:
            print("start ", code.co_name)
            monitor.set_local_events(DEBUGGER_ID, code, monitor.events.LINE | monitor.events.JUMP | monitor.events.PY_YIELD)

    def _on_line(code, line):
        if "test_sys_monitoring.py" in code.co_filename:
            print("on line", code.co_name, line - code.co_firstlineno)

    def _on_jump(code, from_offset, to_offset):
        if "test_sys_monitoring.py" in code.co_filename:
            frame = sys._getframe().f_back
            print("on jump", code.co_name, frame.f_lineno - code.co_firstlineno)

    def _yield_method(code, offset, retval):
        if "test_sys_monitoring.py" in code.co_filename:
            frame = sys._getframe().f_back
            print("on yield", code.co_name, frame.f_lineno - code.co_firstlineno)

    monitor.set_events(DEBUGGER_ID, monitor.events.PY_START | monitor.events.PY_RESUME)

    monitor.register_callback(DEBUGGER_ID, monitor.events.PY_START, _start_method)
    monitor.register_callback(DEBUGGER_ID, monitor.events.PY_RESUME, _start_method)
    monitor.register_callback(DEBUGGER_ID, monitor.events.PY_YIELD, _yield_method)
    monitor.register_callback(DEBUGGER_ID, monitor.events.LINE, _on_line)
    monitor.register_callback(DEBUGGER_ID, monitor.events.JUMP, _on_jump)

    def method():
        a = [1, 2, 3, 4, 5, 6]  # line 1

        # line 3
        def b():  #  line 4
            yield from [j for j in a if j % 2 == 0]  # line 5

        # line 7
        for j in b():  # line 8
            print(j)  # line 9

    method()
    monitor.set_events(DEBUGGER_ID, 0)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest

from tests import code, debug, log, net, test_data
from tests.debug import runners, targets
from tests.patterns import some

pytestmark = pytest.mark.timeout(60)

django_server = net.WebServer(net.get_test_server_port())


class paths:
    django1 = test_data / "django1"
    app_py = django1 / "app.py"
    hello_html = django1 / "templates" / "hello.html"
    bad_html = django1 / "templates" / "bad.html"


class lines:
    app_py = code.get_marked_line_numbers(paths.app_py)


@pytest.fixture
@pytest.mark.parametrize("run", [runners.launch, runners.attach_connect["cli"]])
def start_django(run):
    def start(session, multiprocess=False):
        # No clean way to kill Django server, expect non-zero exit code
        session.expected_exit_code = some.int

        session.config.update({"django": True, "subProcess": bool(multiprocess)})

        args = ["runserver"]
        if not multiprocess:
            args += ["--noreload"]
        args += ["--", str(django_server.port)]

        return run(session, targets.Program(paths.app_py, args=args), cwd=paths.django1)

    return start


@pytest.mark.parametrize("bp_target", ["code", "template"])
def test_django_breakpoint_no_multiproc(start_django, bp_target):
    bp_file, bp_line, bp_name = {
        "code": (paths.app_py, lines.app_py["bphome"], "home"),
        "template": (paths.hello_html, 8, "Django Template"),
    }[bp_target]
    bp_var_content = "Django-Django-Test"

    with debug.Session() as session:
        with start_django(session):
            breakpoints = session.set_breakpoints(bp_file, [bp_line])
            for bp in breakpoints:
                # They'll be verified later on for templates.
                assert bp["verified"] == (bp_target == "code")

        with django_server:
            home_request = django_server.get("/home")

            if bp_target == "template":
                breakpoint_body = session.wait_for_next_event("breakpoint")
                assert breakpoint_body["reason"] == "changed"
                assert breakpoint_body["breakpoint"]["verified"]

            session.wait_for_stop(
                "breakpoint",
                expected_frames=[
                    some.dap.frame(some.dap.source(bp_file), line=bp_line, name=bp_name)
                ],
            )

            var_content = session.get_variable("content")
            assert var_content == some.dict.containing(
                {
                    "name": "content",
                    "type": "str",
                    "value": repr(bp_var_content),
                    "presentationHint": {"attributes": ["rawString"]},
                    "evaluateName": "content",
                    "variablesReference": 0,
                }
            )

            session.request_continue()
            assert bp_var_content in home_request.response_text()


def test_django_template_exception_no_multiproc(start_django):
    with debug.Session() as session:
        with start_django(session):
            session.request(
                "setExceptionBreakpoints", {"filters": ["raised", "uncaught"]}
            )

        with django_server:
            django_server.get("/badtemplate", log_errors=False)
            stop = session.wait_for_stop(
                "exception",
                expected_frames=[
                    some.dap.frame(
                        some.dap.source(paths.bad_html),
                        line=8,
                        name="Django TemplateSyntaxError",
                    )
                ],
            )

            # Will stop once in the plugin
            exception_info = session.request(
                "exceptionInfo", {"threadId": stop.thread_id}
            )
            assert exception_info == some.dict.containing(
                {
                    "exceptionId": some.str.ending_with("TemplateSyntaxError"),
                    "breakMode": "always",
                    "description": some.str.containing("doesnotexist"),
                    "details": some.dict.containing(
                        {
                            "message": some.str.containing("doesnotexist"),
                            "typeName": some.str.ending_with("TemplateSyntaxError"),
                        }
                    ),
                }
            )

            session.request_continue()

            log.info("Exception will be reported again in {0}", paths.app_py)
            session.wait_for_stop("exception")
            session.request_continue()


@pytest.mark.parametrize("exc_type", ["handled", "unhandled"])
def test_django_exception_no_multiproc(start_django, exc_type):
    exc_line = lines.app_py["exc_" + exc_type]

    with debug.Session() as session:
        with start_django(session):
            session.request(
                "setExceptionBreakpoints", {"filters": ["raised", "uncaught"]}
            )

        with django_server:
            django_server.get("/" + exc_type)

            stopped = session.wait_for_stop(
                "exception",
                expected_frames=[
                    some.dap.frame(
                        some.dap.source(paths.app_py),
                        line=exc_line,
                        name="bad_route_" + exc_type,
                    )
                ],
            ).body

            assert stopped == some.dict.containing(
                {
                    "reason": "exception",
                    "text": some.str.ending_with("ArithmeticError"),
                    "description": "Hello",
                }
            )

            exception_info = session.request(
                "exceptionInfo", {"threadId": stopped["threadId"]}
            )

            assert exception_info == {
                "exceptionId": some.str.ending_with("ArithmeticError"),
                "breakMode": "always",
                "description": "Hello",
                "details": {
                    "message": "Hello",
                    "typeName": some.str.ending_with("ArithmeticError"),
                    "source": some.path(paths.app_py),
                    "stackTrace": some.str,
                },
            }

            session.request_continue()


def test_django_breakpoint_multiproc(start_django):
    bp_line = lines.app_py["bphome"]
    bp_var_content = "Django-Django-Test"

    with debug.Session() as parent_session:
        with start_django(parent_session, multiprocess=True):
            parent_session.set_breakpoints(paths.app_py, [bp_line])

        with parent_session.wait_for_next_subprocess() as child_session:
            with child_session.start():
                child_session.set_breakpoints(paths.app_py, [bp_line])

            with django_server:
                home_request = django_server.get("/home")
                child_session.wait_for_stop(
                    "breakpoint",
                    expected_frames=[
                        some.dap.frame(
                            some.dap.source(paths.app_py), line=bp_line, name="home"
                        )
                    ],
                )

                var_content = child_session.get_variable("content")
                assert var_content == some.dict.containing(
                    {
                        "name": "content",
                        "type": "str",
                        "value": repr(bp_var_content),
                        "presentationHint": {"attributes": ["rawString"]},
                        "evaluateName": "content",
                    }
                )

                child_session.request_continue()
                assert bp_var_content in home_request.response_text()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest

from debugpy.common import log
from tests.patterns import some


NONE = None
NAN = float("nan")


def log_repr(x):
    s = repr(x)
    log.info("{0}", s)


VALUES = [
    object(),
    True,
    False,
    0,
    -1,
    -1.0,
    1.23,
    b"abc",
    b"abcd",
    "abc",
    "abcd",
    (),
    (1, 2, 3),
    [],
    [1, 2, 3],
    {},
    {"a": 1, "b": 2},
]


@pytest.mark.parametrize("x", VALUES)
def test_value(x):
    log_repr(some.object)
    assert x == some.object

    log_repr(some.object.equal_to(x))
    assert x == some.object.equal_to(x)

    log_repr(some.object.not_equal_to(x))
    assert x != some.object.not_equal_to(x)

    log_repr(some.object.same_as(x))
    assert x == some.object.same_as(x)

    log_repr(some.thing)
    assert x == some.thing

    log_repr(~some.thing)
    assert x != ~some.thing

    log_repr(~some.object)
    assert x != ~some.object

    log_repr(~some.object | x)
    assert x == ~some.object | x


def test_none():
    assert NONE == some.object
    assert NONE == some.object.equal_to(None)
    assert NONE == some.object.same_as(None)
    assert NONE != some.thing
    assert NONE == some.thing | None


def test_equal():
    assert 123.0 == some.object.equal_to(123)
    assert NAN != some.object.equal_to(NAN)


def test_not_equal():
    assert 123.0 != some.object.not_equal_to(123)
    assert NAN == some.object.not_equal_to(NAN)


def test_same():
    assert 123.0 != some.object.same_as(123)
    assert NAN == some.object.same_as(NAN)


def test_inverse():
    pattern = ~some.object.equal_to(2)
    log_repr(pattern)

    assert pattern == 1
    assert pattern != 2
    assert pattern == 3
    assert pattern == "2"
    assert pattern == NONE


def test_either():
    pattern = some.number | some.str
    log_repr(pattern)
    assert pattern == 123

    pattern = some.str | 123 | some.bool
    log_repr(pattern)
    assert pattern == 123


def test_in_range():
    pattern = some.int.in_range(-5, 5)
    log_repr(pattern)

    assert all([pattern == x for x in range(-5, 5)])
    assert pattern != -6
    assert pattern != 5


def test_str():
    log_repr(some.str)
    assert some.str == "abc"
    assert b"abc" != some.str


def test_matching():
    pattern = some.str.matching(r".(b+).")
    log_repr(pattern)
    assert pattern == "abbbc"

    pattern = some.str.matching(r"bbb")
    log_repr(pattern)
    assert pattern != "abbbc"

    pattern = some.bytes.matching(rb".(b+).")
    log_repr(pattern)
    assert pattern == b"abbbc"

    pattern = some.bytes.matching(rb"bbb")
    log_repr(pattern)
    assert pattern != b"abbbc"


def test_starting_with():
    pattern = some.str.starting_with("aa")
    log_repr(pattern)
    assert pattern == "aabbbb"
    assert pattern != "bbbbaa"
    assert pattern != "bbaabb"
    assert pattern != "ababab"

    pattern = some.bytes.starting_with(b"aa")
    log_repr(pattern)
    assert pattern == b"aabbbb"
    assert pattern != b"bbbbaa"
    assert pattern != b"bbaabb"
    assert pattern != b"ababab"


def test_ending_with():
    pattern = some.str.ending_with("aa")
    log_repr(pattern)
    assert pattern == "bbbbaa"
    assert pattern == "bb\nbb\naa"
    assert pattern != "aabbbb"
    assert pattern != "bbaabb"
    assert pattern != "ababab"

    pattern = some.bytes.ending_with(b"aa")
    log_repr(pattern)
    assert pattern == b"bbbbaa"
    assert pattern == b"bb\nbb\naa"
    assert pattern != b"aabbbb"
    assert pattern != b"bbaabb"
    assert pattern != b"ababab"


def test_containing():
    pattern = some.str.containing("aa")
    log_repr(pattern)
    assert pattern == "aabbbb"
    assert pattern == "bbbbaa"
    assert pattern == "bbaabb"
    assert pattern == "bb\naa\nbb"
    assert pattern != "ababab"

    pattern = some.bytes.containing(b"aa")
    log_repr(pattern)
    assert pattern == b"aabbbb"
    assert pattern == b"bbbbaa"
    assert pattern == b"bbaabb"
    assert pattern == b"bb\naa\nbb"
    assert pattern != b"ababab"


def test_list():
    assert [1, 2, 3] == [1, some.thing, 3]
    assert [1, 2, 3, 4] != [1, some.thing, 4]

    assert [1, 2, 3, 4] == some.list.containing(1)
    assert [1, 2, 3, 4] == some.list.containing(2)
    assert [1, 2, 3, 4] == some.list.containing(3)
    assert [1, 2, 3, 4] == some.list.containing(4)
    assert [1, 2, 3, 4] == some.list.containing(1, 2)
    assert [1, 2, 3, 4] == some.list.containing(2, 3)
    assert [1, 2, 3, 4] == some.list.containing(3, 4)
    assert [1, 2, 3, 4] == some.list.containing(1, 2, 3)
    assert [1, 2, 3, 4] == some.list.containing(2, 3, 4)
    assert [1, 2, 3, 4] == some.list.containing(1, 2, 3, 4)

    assert [1, 2, 3, 4] != some.list.containing(5)
    assert [1, 2, 3, 4] != some.list.containing(1, 3)
    assert [1, 2, 3, 4] != some.list.containing(1, 2, 4)
    assert [1, 2, 3, 4] != some.list.containing(2, 3, 5)


def test_dict():
    pattern = {"a": some.thing, "b": 2}
    log_repr(pattern)
    assert pattern == {"a": 1, "b": 2}

    pattern = some.dict.containing({"a": 1})
    log_repr(pattern)
    assert pattern == {"a": 1, "b": 2}


def test_such_that():
    pattern = some.thing.such_that(lambda x: x != 1)
    log_repr(pattern)

    assert 0 == pattern
    assert 1 != pattern
    assert 2 == pattern


def test_error():
    log_repr(some.error)
    assert some.error == Exception("error!")
    assert some.error != {}


def test_recursive():
    pattern = some.dict.containing(
        {
            "dict": some.dict.containing({"int": some.int.in_range(100, 200)}),
            "list": [None, ~some.error, some.number | some.str],
        }
    )

    log_repr(pattern)

    assert pattern == {
        "list": [None, False, 123],
        "bool": True,
        "dict": {"int": 123, "str": "abc"},
    }



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest

from tests import debug
from tests.debug import runners, targets
from tests.patterns import some


@pytest.mark.parametrize("breakpoint", ["breakpoint", ""])
@pytest.mark.parametrize("run", runners.all_launch)
@pytest.mark.parametrize("target", targets.all_named)
def test_stop_on_entry(pyfile, run, target, breakpoint):
    @pyfile
    def code_to_debug():
        from debuggee import backchannel  # @bp

        backchannel.send("done")

    with debug.Session() as session:
        session.config["stopOnEntry"] = True

        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            if breakpoint:
                session.set_breakpoints(code_to_debug, all)

        if breakpoint:
            stop = session.wait_for_stop(
                "breakpoint", expected_frames=[some.dap.frame(code_to_debug, "bp")]
            )
            session.request("next", {"threadId": stop.thread_id})
            stop = session.wait_for_stop(
                "step", expected_frames=[some.dap.frame(code_to_debug, 3)]
            )
        else:
            session.wait_for_stop(
                "entry", expected_frames=[some.dap.frame(code_to_debug, 1)]
            )

        session.request_continue()
        assert backchannel.receive() == "done"
